import numpy as np
from astropy.table import QTable
import astropy.units as u
from tqdm import tqdm
import operator

from .cuts import evaluate_binned_cut, calculate_percentile_cut
from .sensitivity import calculate_sensitivity, estimate_background
from .binning import create_histogram_table, bin_center


__all__ = [
    "optimize_gh_cut",
]


def optimize_gh_cut(
    signal,
    background,
    reco_energy_bins,
    gh_cut_efficiencies,
    theta_cuts,
    op=operator.ge,
    fov_offset_min=0 * u.deg,
    fov_offset_max=1 * u.deg,
    alpha=1.0,
    progress=True,
    **kwargs
):
    """
    Optimize the gh-score cut in every energy bin of reconstructed energy
    for best sensitivity.

    This procedure is EventDisplay-like, since it only applies a
    pre-computed theta cut and then optimizes only the gamma/hadron separation
    cut.

    Parameters
    ----------
    signal: astropy.table.QTable
        event list of simulated signal events.
        Required columns are `theta`, `reco_energy`, 'weight', `gh_score`
        No directional (theta) or gamma/hadron cut should already be applied.
    background: astropy.table.QTable
        event list of simulated background events.
        Required columns are `reco_source_fov_offset`, `reco_energy`,
        'weight', `gh_score`.
        No directional (theta) or gamma/hadron cut should already be applied.
    reco_energy_bins: astropy.units.Quantity[energy]
        Bins in reconstructed energy to use for sensitivity computation
    gh_cut_efficiencies: np.ndarray[float, ndim=1]
        The cut efficiencies to scan for best sensitivity.
    theta_cuts: astropy.table.QTable
        cut definition of the energy dependent theta cut,
        e.g. as created by ``calculate_percentile_cut``
    op: comparison function with signature f(a, b) -> bool
        The comparison function to use for the gamma hadron score.
        Returning true means an event passes the cut, so is not discarded.
        E.g. for gammaness-like score, use `operator.ge` (>=) and for a
        hadroness-like score use `operator.le` (<=).
    fov_offset_min: astropy.units.Quantity[angle]
        Minimum distance from the fov center for background events to be taken into account
    fov_offset_max: astropy.units.Quantity[angle]
        Maximum distance from the fov center for background events to be taken into account
    alpha: float
        Size ratio of off region / on region. Will be used to
        scale the background rate.
    progress: bool
        If True, show a progress bar during cut optimization
    **kwargs are passed to ``calculate_sensitivity``
    """

    # we apply each cut for all reco_energy_bins globally, calculate the
    # sensitivity and then lookup the best sensitivity for each
    # bin independently

    signal_selected_theta = evaluate_binned_cut(
        signal['theta'], signal['reco_energy'], theta_cuts,
        op=operator.le,
    )

    sensitivities = []
    gh_cuts = []
    for efficiency in tqdm(gh_cut_efficiencies, disable=not progress):

        # calculate necessary percentile needed for
        # ``calculate_percentile_cut`` with the correct efficiency.
        # Depends on the operator, since we need to invert the
        # efficiency if we compare using >=, since percentile is
        # defines as <=.
        if op(-1, 1): # if operator behaves like "<=", "<" etc:
            percentile = 100 * efficiency
            fill_value = signal['gh_score'].min()
        else: # operator behaves like ">=", ">"
            percentile = 100 * (1 - efficiency)
            fill_value = signal['gh_score'].max()

        gh_cut = calculate_percentile_cut(
            signal['gh_score'], signal['reco_energy'],
            bins=reco_energy_bins,
            fill_value=fill_value, percentile=percentile,
        )
        gh_cuts.append(gh_cut)

        # apply the current cut
        signal_selected = evaluate_binned_cut(
            signal["gh_score"], signal["reco_energy"], gh_cut, op,
        ) & signal_selected_theta

        background_selected = evaluate_binned_cut(
            background["gh_score"], background["reco_energy"], gh_cut, op,
        )

        # create the histograms
        signal_hist = create_histogram_table(
            signal[signal_selected], reco_energy_bins, "reco_energy"
        )

        background_hist = estimate_background(
            events=background[background_selected],
            reco_energy_bins=reco_energy_bins,
            theta_cuts=theta_cuts,
            alpha=alpha,
            fov_offset_min=fov_offset_min,
            fov_offset_max=fov_offset_max,
        )

        sensitivity = calculate_sensitivity(
            signal_hist, background_hist, alpha=alpha,
            **kwargs,
        )
        sensitivities.append(sensitivity)

    best_cut_table = QTable()
    best_cut_table["low"] = reco_energy_bins[0:-1]
    best_cut_table["center"] = bin_center(reco_energy_bins)
    best_cut_table["high"] = reco_energy_bins[1:]
    best_cut_table["cut"] = np.nan

    best_sensitivity = sensitivities[0].copy()
    for bin_id in range(len(reco_energy_bins) - 1):
        sensitivities_bin = [s["relative_sensitivity"][bin_id] for s in sensitivities]

        if not np.all(np.isnan(sensitivities_bin)):
            # nanargmin won't return the index of nan entries
            best = np.nanargmin(sensitivities_bin)
        else:
            # if all are invalid, just use the first one
            best = 0

        best_sensitivity[bin_id] = sensitivities[best][bin_id]
        best_cut_table["cut"][bin_id] = gh_cuts[best]["cut"][bin_id]

    return best_sensitivity, best_cut_table



import logging

from astropy.table import QTable, unique
import astropy.units as u

from ..simulations import SimulatedEventsInfo


log = logging.getLogger(__name__)


COLUMN_MAP = {
    "obs_id": "OBS_ID",
    "event_id": "EVENT_ID",
    "true_energy": "MC_ENERGY",
    "reco_energy": "ENERGY",
    "true_alt": "MC_ALT",
    "true_az": "MC_AZ",
    "pointing_alt": "PNT_ALT",
    "pointing_az": "PNT_AZ",
    "reco_alt": "ALT",
    "reco_az": "AZ",
    "gh_score": "GH_MVA",
    "multiplicity": "MULTIP",
}


def read_eventdisplay_fits(infile, use_histogram=True):
    """
    Read a DL2 FITS file as produced by the EventDisplay DL2 converter
    from ROOT files:
    https://github.com/Eventdisplay/Converters/blob/master/DL2/generate_DL2_file.py

    Parameters
    ----------
    infile : str or pathlib.Path
        Path to the input fits file
    use_histogram : bool
        If True, use number of simulated events from histogram provided in fits file,
        if False, estimate this number from the unique run_id, pointing direction
        combinations and the number of events per run in the run header.
        This will fail e.g. for protons with cuts already applied, since many
        runs will have 0 events surviving cuts.

    Returns
    -------
    events: astropy.QTable
        Astropy Table object containing the reconstructed events information.
    simulated_events: ``~pyirf.simulations.SimulatedEventsInfo``
    """
    log.debug(f"Reading {infile}")
    events = QTable.read(infile, hdu="EVENTS")
    sim_events = QTable.read(infile, hdu="SIMULATED EVENTS")
    run_header = QTable.read(infile, hdu="RUNHEADER")[0]

    for new, old in COLUMN_MAP.items():
        events.rename_column(old, new)

    n_runs = len(unique(events[['obs_id', 'pointing_az', 'pointing_alt']]))
    log.info(f"Estimated number of runs from obs ids and pointing position: {n_runs}")

    n_showers_guessed = n_runs * run_header["num_use"] * run_header["num_showers"]
    n_showers_hist = int(sim_events["EVENTS"].sum())

    if use_histogram:
        n_showers = n_showers_hist
    else:
        n_showers = n_showers_guessed

    log.debug("Number of events histogram: %d", n_showers_hist)
    log.debug("Number of events from n_runs and run header: %d", n_showers_guessed)
    log.debug("Using number of events from %s", "histogram" if use_histogram else "guess")

    sim_info = SimulatedEventsInfo(
        n_showers=n_showers,
        energy_min=u.Quantity(run_header["E_range"][0], u.TeV),
        energy_max=u.Quantity(run_header["E_range"][1], u.TeV),
        max_impact=u.Quantity(run_header["core_range"][1], u.m),
        spectral_index=run_header["spectral_index"],
        viewcone=u.Quantity(run_header["viewcone"][1], u.deg),
    )

    return events, sim_info



import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u
from scipy.stats import norm

from pyirf.spectral import (
    DAMPE_P_He_SPECTRUM,
    IRFDOC_ELECTRON_SPECTRUM,
    IRFDOC_PROTON_SPECTRUM,
    PDG_ALL_PARTICLE,
    CRAB_HEGRA,
    CRAB_MAGIC_JHEAP2015,
    POINT_SOURCE_FLUX_UNIT,
    DIFFUSE_FLUX_UNIT,
)


cr_spectra = {
    "PDG All Particle Spectrum": PDG_ALL_PARTICLE,
    "ATIC Proton Fit (from IRF Document)": IRFDOC_PROTON_SPECTRUM,
    "DAMPE p + He Table Interpolation": DAMPE_P_He_SPECTRUM
}


if __name__ == "__main__":

    energy = np.geomspace(0.001, 300, 1000) * u.TeV

    plt.figure(constrained_layout=True)
    plt.title("Crab Nebula Flux")
    plt.plot(
        energy.to_value(u.TeV),
        CRAB_HEGRA(energy).to_value(POINT_SOURCE_FLUX_UNIT),
        label="HEGRA",
    )
    plt.plot(
        energy.to_value(u.TeV),
        CRAB_MAGIC_JHEAP2015(energy).to_value(POINT_SOURCE_FLUX_UNIT),
        label="MAGIC JHEAP 2015",
    )

    plt.legend()
    plt.xscale("log")
    plt.yscale("log")
    plt.xlabel("E / TeV")
    plt.ylabel(f'Flux / ({POINT_SOURCE_FLUX_UNIT.to_string("latex")})')

    plt.figure(constrained_layout=True)
    plt.title("Cosmic Ray Flux")

    for label, spectrum in cr_spectra.items():
        unit = energy.unit ** 2 * DIFFUSE_FLUX_UNIT
        plt.plot(
            energy.to_value(u.TeV),
            (spectrum(energy) * energy ** 2).to_value(unit),
            label=label,
        )

    plt.legend()
    plt.xscale("log")
    plt.yscale("log")
    plt.xlabel(r"$E \,\,/\,\, \mathrm{TeV}$")
    plt.ylabel(rf'$E^{2} \cdot \Phi \,\,/\,\,$ ({unit.to_string("latex")})')

    energy = np.geomspace(0.006, 10, 1000) * u.TeV
    plt.figure(constrained_layout=True)
    plt.title("Electron Flux")

    unit = u.TeV ** 2 / u.m ** 2 / u.s / u.sr
    plt.plot(
        energy.to_value(u.TeV),
        (energy ** 3 * IRFDOC_ELECTRON_SPECTRUM(energy)).to_value(unit),
        label="IFAE 2013 (from IRF Document)",
    )

    plt.legend()
    plt.xscale("log")
    plt.xlim(5e-3, 10)
    plt.ylim(1e-5, 0.25e-3)
    plt.xlabel(r"$E \,\,/\,\, \mathrm{TeV}$")
    plt.ylabel(rf'$E^3 \cdot \Phi \,\,/\,\,$ ({unit.to_string("latex")})')
    plt.grid()

    plt.show()



"""
Simple wrapper around scipy.interpolate.griddata to interpolate parametrized quantities
"""
from scipy.interpolate import griddata

from .base_interpolators import ParametrizedInterpolator

__all__ = ["GridDataInterpolator"]


class GridDataInterpolator(ParametrizedInterpolator):
    def __init__(self, grid_points, params):
        """GridDataInterpolator constructor

        Parameters
        ----------
        grid_points: np.ndarray, shape=(N, O)
            Grid points at which interpolation templates exist
        params: np.ndarray, shape=(N, ...)
            Structured array of corresponding parameter values at each
            point in grid_points.
            First dimesion has to correspond to number of grid_points

        Raises
        ------
        TypeError:
            When params is not a np.ndarray
        ValueError:
            When number of points grid_points and params is not matching
        """
        super().__init__(grid_points, params)

    def interpolate(self, target_point, **kwargs):
        """
        Wrapper around scipy.interpolate.griddata [1]

        Parameters
        ----------
        target_point: np.ndarray, shape=(1, O)
            Target point for interpolation
        **kwargs:
            Passed to scipy.interpolate.griddata [1]

        Returns
        -------
        interpolant: np.ndarray

        References
        ----------
        .. [1] Scipy Documentation, scipy.interpolate.griddata
               https://docs.scipy.org/doc/scipy/reference/generated
        """
        interpolant = griddata(
            self.grid_points, self.params, target_point, **kwargs
        ).squeeze()

        return interpolant.reshape(1, *self.params.shape[1:])



"""
Collection of interpolation methods
"""

from .base_interpolators import (
    BaseInterpolator,
    BinnedInterpolator,
    ParametrizedInterpolator,
)
from .griddata_interpolator import GridDataInterpolator
from .interpolate_irfs import (
    interpolate_effective_area_per_energy_and_fov,
    interpolate_energy_dispersion,
    interpolate_psf_table,
    interpolate_rad_max,
)
from .quantile_interpolator import QuantileInterpolator

__all__ = [
    "BaseInterpolator",
    "BinnedInterpolator",
    "GridDataInterpolator",
    "ParametrizedInterpolator",
    "QuantileInterpolator",
    "interpolate_effective_area_per_energy_and_fov",
    "interpolate_energy_dispersion",
    "interpolate_psf_table",
    "interpolate_rad_max",
]



"""Tests for base interpolator classes"""
import re

import numpy as np
import pytest


def test_BaseInterpolator_datastructure_checks():
    """Test checks performed to enforce correct data structure"""
    from pyirf.interpolation.base_interpolators import BaseInterpolator

    grid_points1D_good = np.array([1, 2, 3])
    grid_points1D_bad = [1, 2, 3]
    target1D_bad = 1

    grid_points2D_good = np.array([[1, 1], [2, 1], [1.5, 1.5]])
    grid_points2D_bad_str = np.array([["1", "1"], ["2", "1"], ["1.5", "1.5"]])
    grid_points2D_bad_obj = np.array([[1, 1], [2, 1], [1.5]], dtype="object")
    target2D_bad = [1.25, 1.25]

    with pytest.raises(TypeError):  # Abstract class, cannot be instantiated
        BaseInterpolator(grid_points1D_good)

    class DummyBaseInterpolator(BaseInterpolator):
        def interpolate(self, target_point, **kwargs):
            raise NotImplementedError

    with pytest.raises(TypeError, match="Input grid_points is not a numpy array."):
        DummyBaseInterpolator(grid_points1D_bad)

    with pytest.raises(
        TypeError, match="Input grid_points array cannot be of dtype object."
    ):
        DummyBaseInterpolator(grid_points2D_bad_obj)

    with pytest.raises(
        TypeError, match="Input grid_points dtype incompatible with float."
    ):
        DummyBaseInterpolator(grid_points2D_bad_str)

    with pytest.raises(TypeError, match="Target point is not a numpy array."):
        Interp = DummyBaseInterpolator(grid_points1D_good)
        Interp(target1D_bad)

    with pytest.raises(TypeError, match="Target point is not a numpy array."):
        Interp = DummyBaseInterpolator(grid_points2D_good)
        Interp(target2D_bad)


def test_BaseInterpolator_sanity_checks():
    """Test checks performed to enforce correct data values"""
    from pyirf.interpolation.base_interpolators import BaseInterpolator

    grid_points1D_good = np.array([1, 2, 3])
    target1D_inGrid = np.array([1.5])
    target1D_outofGrid = np.array([0.5])

    grid_points2D_good = np.array([[1, 1], [2, 1], [1.5, 1.5]])
    grid_points2D_toofew = np.array([[1, 1], [2, 1]])
    target2D_inGrid = np.array([1.25, 1.25])
    target2D_outofGrid = np.array([0.5, 0.5])
    target2D_twopoints = np.array([[1.25, 1.25], [1.25, 1.25]])

    class DummyBaseInterpolator(BaseInterpolator):
        def interpolate(self, target_point, **kwargs):
            raise NotImplementedError

    with pytest.raises(
        ValueError,
        match="Target point outside grids convex hull and no extrapolator given.",
    ):
        interp = DummyBaseInterpolator(grid_points1D_good)
        interp(target1D_outofGrid)

    with pytest.raises(ValueError, match="To few points for grid dimension"):
        DummyBaseInterpolator(grid_points2D_toofew)

    with pytest.raises(ValueError, match="Only one target_point per call supported."):
        interp = DummyBaseInterpolator(grid_points2D_good)
        interp(target2D_twopoints)

    with pytest.raises(
        ValueError,
        match="Target point outside grids convex hull and no extrapolator given.",
    ):
        interp = DummyBaseInterpolator(grid_points2D_good)
        interp(target2D_outofGrid)

    with pytest.raises(
        ValueError, match="Missmatch between target-point and grid dimension."
    ):
        interp = DummyBaseInterpolator(grid_points2D_good)
        interp(target1D_inGrid)

    with pytest.raises(NotImplementedError):
        # Everything ok but interpolate not implemented
        interp = DummyBaseInterpolator(grid_points1D_good)
        interp(target1D_inGrid)

    with pytest.raises(NotImplementedError):
        # Everything ok but interpolate not implemented
        interp = DummyBaseInterpolator(grid_points2D_good)
        interp(target2D_inGrid)


def test_BaseInterpolator_extrapolation():
    """Test extrapolators are correctly passed"""
    from pyirf.interpolation.base_interpolators import BaseInterpolator

    def dummy_extrapolator(target_point):
        return np.sum(target_point)

    class DummyBaseInterpolator(BaseInterpolator):
        def interpolate(self, target_point, **kwargs):
            raise NotImplementedError

    grid_points1D_good = np.array([1, 2, 3])
    target1D_outofGrid = np.array([0.5])

    grid_points2D_good = np.array([[1, 1], [2, 1], [1.5, 1.5]])
    target2D_outofGrid = np.array([0.5, 0.5])

    # 1D target out of grid but extrapolator given
    interp1D = DummyBaseInterpolator(grid_points1D_good)
    assert interp1D(target1D_outofGrid, extrapolator=dummy_extrapolator) == np.sum(
        target1D_outofGrid
    )

    # 2D target out of grid but extrapolator given
    interp2D = DummyBaseInterpolator(grid_points2D_good)
    assert interp2D(target2D_outofGrid, extrapolator=dummy_extrapolator) == np.sum(
        target2D_outofGrid
    )


def test_ParametrizedInterpolator():
    """Test ParametrizedInterpolator initialization and sanity checks"""
    from pyirf.interpolation.base_interpolators import ParametrizedInterpolator

    grid_points = np.array([1, 2, 3])
    params_good = np.array([[1], [2], [3]])
    params_shape_missmatch = np.array([[1], [2]])

    with pytest.raises(TypeError):  # Abstract class, cannot be instantiated
        ParametrizedInterpolator(grid_points, params_good)

    class DummyParametrizedInterpolator(ParametrizedInterpolator):
        def interpolate(self, target_point, **kwargs):
            raise NotImplementedError

    with pytest.raises(TypeError, match="Input params is not a numpy array."):
        DummyParametrizedInterpolator(grid_points, params_good.tolist())

    with pytest.raises(
        ValueError,
        match="Shape missmatch, number of grid_points and rows in params not matching.",
    ):
        DummyParametrizedInterpolator(grid_points, params_shape_missmatch)

    interp = DummyParametrizedInterpolator(grid_points, params_good)

    with pytest.raises(NotImplementedError):
        # Everything ok but interpolate not implemented
        interp(np.array([1]))


def test_BinnedInterpolator():
    """Test BinnedInterpolator initialization and sanity checks"""
    from pyirf.interpolation.base_interpolators import BinnedInterpolator

    grid_points = np.array([1, 2, 3])
    bin_edges = np.linspace(-1, 1, 11)
    bin_content_good_shape = np.ones(shape=(len(grid_points), len(bin_edges) - 1))
    bin_content_bad_nhist = np.ones(shape=(len(grid_points) - 1, len(bin_edges) - 1))
    bin_content_bad_nbins = np.ones(shape=(len(grid_points), len(bin_edges)))

    with pytest.raises(TypeError):  # Abstract class, cannot be instantiated
        BinnedInterpolator(grid_points, bin_edges, bin_content_good_shape)

    class DummyBinnedInterpolator(BinnedInterpolator):
        def interpolate(self, target_point, **kwargs):
            raise NotImplementedError

    with pytest.raises(TypeError, match="Input bin_edges is not a numpy array."):
        DummyBinnedInterpolator(grid_points, bin_edges.tolist(), bin_content_good_shape)

    with pytest.raises(TypeError, match="Input bin_contents is not a numpy array."):
        DummyBinnedInterpolator(grid_points, bin_edges, bin_content_good_shape.tolist())

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Shape missmatch, number of grid_points (3) and "
            "number of histograms in bin_contents (2) not matching."
        ),
    ):
        DummyBinnedInterpolator(grid_points, bin_edges, bin_content_bad_nhist)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Shape missmatch, bin_edges (10 bins) "
            "and bin_contents (11 bins) not matching."
        ),
    ):
        DummyBinnedInterpolator(grid_points, bin_edges, bin_content_bad_nbins)

    interp = DummyBinnedInterpolator(grid_points, bin_edges, bin_content_good_shape)

    with pytest.raises(NotImplementedError):
        # Everything ok but interpolate not implemented
        interp(np.array([1]))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))
from pyirf import __version__

# -- Project information -----------------------------------------------------

project = "pyirf"
copyright = "2020, Maximilian Nöthe, Michele Peresano, Thomas Vuillaume"
author = "Maximilian Nöthe, Michele Peresano, Thomas Vuillaume"

# The full version, including alpha/beta/rc tags
version = __version__

# -- General configuration ---------------------------------------------------

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]
# source_suffix = '.rst'

# The master toctree document.
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "numpydoc",
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "IPython.sphinxext.ipython_console_highlighting",
]

# nbsphinx
# nbsphinx_execute = "never"
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'png', }",
    "--InlineBackend.rc={'figure.dpi': 300}",
]

numpydoc_show_class_members = False
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "changes"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    'canonical_url': 'https://cta-observatory.github.io/pyirf',
    'display_version': True,
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.7", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "astropy": ("https://docs.astropy.org/en/latest/", None),
    "gammapy": ("https://docs.gammapy.org/0.18/", None),
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



import numpy as np

from .utils import is_scalar

__all__ = ["li_ma_significance"]


def li_ma_significance(n_on, n_off, alpha=0.2):
    """
    Calculate the Li & Ma significance.

    Formula (17) in https://doi.org/10.1086/161295

    This functions returns 0 significance when n_on < alpha * n_off
    instead of the negative sensitivities that would result from naively
    evaluating the formula.

    Parameters
    ----------
    n_on: integer or array like
        Number of events for the on observations
    n_off: integer or array like
        Number of events for the off observations
    alpha: float
        Ratio between the on region and the off region size or obstime.

    Returns
    -------
    s_lima: float or array
        The calculated significance
    """

    scalar = is_scalar(n_on)

    # Cast everything into float64 to avoid numeric instabilties
    # when multiplying very small and very big numbers to get t1 and t2
    n_on = np.array(n_on, copy=False, ndmin=1, dtype=np.float64)
    n_off = np.array(n_off, copy=False, ndmin=1, dtype=np.float64)
    alpha = np.float64(alpha)

    with np.errstate(divide="ignore", invalid="ignore"):
        p_on = n_on / (n_on + n_off)
        p_off = n_off / (n_on + n_off)

        t1 = n_on * np.log(((1 + alpha) / alpha) * p_on)
        t2 = n_off * np.log((1 + alpha) * p_off)

        # lim x+->0  (x log(x)) = 0
        t1[n_on == 0] = 0
        t2[n_off == 0] = 0

        ts = t1 + t2

        significance = np.sqrt(ts * 2)

    significance[n_on < (alpha * n_off)] = 0

    if scalar:
        return significance[0]

    return significance



"""
Example for using pyirf to calculate IRFS and sensitivity from EventDisplay DL2 fits
files produced from the root output by this script:

https://github.com/Eventdisplay/Converters/blob/master/DL2/generate_DL2_file.py
"""
import logging
import operator

import numpy as np
from astropy import table
import astropy.units as u
from astropy.io import fits

from pyirf.io.eventdisplay import read_eventdisplay_fits
from pyirf.binning import (
    bin_center, create_bins_per_decade,
    add_overflow_bins,
    create_histogram_table,
)
from pyirf.cuts import calculate_percentile_cut, evaluate_binned_cut
from pyirf.sensitivity import calculate_sensitivity, estimate_background
from pyirf.utils import calculate_theta, calculate_source_fov_offset
from pyirf.benchmarks import energy_bias_resolution, angular_resolution

from pyirf.spectral import (
    calculate_event_weights,
    PowerLaw,
    CRAB_HEGRA,
    IRFDOC_PROTON_SPECTRUM,
    IRFDOC_ELECTRON_SPECTRUM,
)
from pyirf.cut_optimization import optimize_gh_cut

from pyirf.irf import (
    effective_area_per_energy,
    energy_dispersion,
    psf_table,
    background_2d,
)

from pyirf.io import (
    create_aeff2d_hdu,
    create_psf_table_hdu,
    create_energy_dispersion_hdu,
    create_rad_max_hdu,
    create_background_2d_hdu,
)


log = logging.getLogger("pyirf")


T_OBS = 50 * u.hour

# scaling between on and off region.
# Make off region 5 times larger than on region for better
# background statistics
ALPHA = 0.2

# Radius to use for calculating bg rate
MAX_BG_RADIUS = 1 * u.deg
MAX_GH_CUT_EFFICIENCY = 0.8
GH_CUT_EFFICIENCY_STEP = 0.01

# gh cut used for first calculation of the binned theta cuts
INITIAL_GH_CUT_EFFICENCY = 0.4

particles = {
    "gamma": {
        "file": "data/gamma_onSource.S.3HB9-FD_ID0.eff-0.fits.gz",
        "target_spectrum": CRAB_HEGRA,
    },
    "proton": {
        "file": "data/proton_onSource.S.3HB9-FD_ID0.eff-0.fits.gz",
        "target_spectrum": IRFDOC_PROTON_SPECTRUM,
    },
    "electron": {
        "file": "data/electron_onSource.S.3HB9-FD_ID0.eff-0.fits.gz",
        "target_spectrum": IRFDOC_ELECTRON_SPECTRUM,
    },
}


def main():
    logging.basicConfig(level=logging.INFO)
    logging.getLogger("pyirf").setLevel(logging.DEBUG)

    for particle_type, p in particles.items():
        log.info(f"Simulated {particle_type.title()} Events:")
        p["events"], p["simulation_info"] = read_eventdisplay_fits(p["file"])
        p["events"]["particle_type"] = particle_type

        p["simulated_spectrum"] = PowerLaw.from_simulation(p["simulation_info"], T_OBS)
        p["events"]["weight"] = calculate_event_weights(
            p["events"]["true_energy"], p["target_spectrum"], p["simulated_spectrum"]
        )
        for prefix in ('true', 'reco'):
            k = f"{prefix}_source_fov_offset"
            p["events"][k] = calculate_source_fov_offset(p["events"], prefix=prefix)

        # calculate theta / distance between reco and assuemd source positoin
        # we handle only ON observations here, so the assumed source pos
        # is the pointing position
        p["events"]["theta"] = calculate_theta(
            p["events"],
            assumed_source_az=p["events"]["pointing_az"],
            assumed_source_alt=p["events"]["pointing_alt"],
        )
        log.info(p["simulation_info"])
        log.info("")

    gammas = particles["gamma"]["events"]
    # background table composed of both electrons and protons
    background = table.vstack(
        [particles["proton"]["events"], particles["electron"]["events"]]
    )

    INITIAL_GH_CUT = np.quantile(gammas['gh_score'], (1 - INITIAL_GH_CUT_EFFICENCY))
    log.info(f"Using fixed G/H cut of {INITIAL_GH_CUT} to calculate theta cuts")

    # event display uses much finer bins for the theta cut than
    # for the sensitivity
    theta_bins = add_overflow_bins(
        create_bins_per_decade(10 ** (-1.9) * u.TeV, 10 ** 2.3005 * u.TeV, 50)
    )
    # same bins as event display uses
    sensitivity_bins = add_overflow_bins(
        create_bins_per_decade(
            10 ** -1.9 * u.TeV, 10 ** 2.31 * u.TeV, bins_per_decade=5
        )
    )

    # theta cut is 68 percent containmente of the gammas
    # for now with a fixed global, unoptimized score cut
    # the cut is calculated in the same bins as the sensitivity,
    # but then interpolated to 10x the resolution.
    mask_theta_cuts = gammas["gh_score"] >= INITIAL_GH_CUT
    theta_cuts_coarse = calculate_percentile_cut(
        gammas["theta"][mask_theta_cuts],
        gammas["reco_energy"][mask_theta_cuts],
        bins=sensitivity_bins,
        min_value=0.05 * u.deg,
        fill_value=0.32 * u.deg,
        max_value=0.32 * u.deg,
        percentile=68,
    )

    # interpolate to 50 bins per decade
    theta_center = bin_center(theta_bins)
    inter_center = bin_center(sensitivity_bins)
    theta_cuts = table.QTable({
        "low": theta_bins[:-1],
        "high": theta_bins[1:],
        "center": theta_center,
        "cut": np.interp(np.log10(theta_center / u.TeV), np.log10(inter_center / u.TeV), theta_cuts_coarse['cut']),
    })


    log.info("Optimizing G/H separation cut for best sensitivity")
    gh_cut_efficiencies = np.arange(
        GH_CUT_EFFICIENCY_STEP,
        MAX_GH_CUT_EFFICIENCY + GH_CUT_EFFICIENCY_STEP / 2,
        GH_CUT_EFFICIENCY_STEP
    )
    sensitivity, gh_cuts = optimize_gh_cut(
        gammas,
        background,
        reco_energy_bins=sensitivity_bins,
        gh_cut_efficiencies=gh_cut_efficiencies,
        op=operator.ge,
        theta_cuts=theta_cuts,
        alpha=ALPHA,
        fov_offset_max=MAX_BG_RADIUS,
    )

    # now that we have the optimized gh cuts, we recalculate the theta
    # cut as 68 percent containment on the events surviving these cuts.
    log.info('Recalculating theta cut for optimized GH Cuts')
    for tab in (gammas, background):
        tab["selected_gh"] = evaluate_binned_cut(
            tab["gh_score"], tab["reco_energy"], gh_cuts, operator.ge
        )

    gammas["selected_theta"] = evaluate_binned_cut(
        gammas["theta"], gammas["reco_energy"], theta_cuts, operator.le
    )
    gammas["selected"] = gammas["selected_theta"] & gammas["selected_gh"]

    # scale relative sensitivity by Crab flux to get the flux sensitivity
    spectrum = particles['gamma']['target_spectrum']
    sensitivity["flux_sensitivity"] = (
        sensitivity["relative_sensitivity"] * spectrum(sensitivity['reco_energy_center'])
    )

    log.info('Calculating IRFs')
    hdus = [
        fits.PrimaryHDU(),
        fits.BinTableHDU(sensitivity, name="SENSITIVITY"),
        fits.BinTableHDU(theta_cuts, name="THETA_CUTS"),
        fits.BinTableHDU(gh_cuts, name="GH_CUTS"),
    ]

    masks = {
        "": gammas["selected"],
        "_NO_CUTS": slice(None),
        "_ONLY_GH": gammas["selected_gh"],
        "_ONLY_THETA": gammas["selected_theta"],
    }

    # binnings for the irfs
    true_energy_bins = add_overflow_bins(
        create_bins_per_decade(10 ** -1.9 * u.TeV, 10 ** 2.31 * u.TeV, 10)
    )
    reco_energy_bins = add_overflow_bins(
        create_bins_per_decade(10 ** -1.9 * u.TeV, 10 ** 2.31 * u.TeV, 5)
    )
    fov_offset_bins = [0, 0.5] * u.deg
    source_offset_bins = np.arange(0, 1 + 1e-4, 1e-3) * u.deg
    energy_migration_bins = np.geomspace(0.2, 5, 200)

    for label, mask in masks.items():
        effective_area = effective_area_per_energy(
            gammas[mask],
            particles["gamma"]["simulation_info"],
            true_energy_bins=true_energy_bins,
        )
        hdus.append(
            create_aeff2d_hdu(
                effective_area[..., np.newaxis],  # add one dimension for FOV offset
                true_energy_bins,
                fov_offset_bins,
                extname="EFFECTIVE_AREA" + label,
            )
        )
        edisp = energy_dispersion(
            gammas[mask],
            true_energy_bins=true_energy_bins,
            fov_offset_bins=fov_offset_bins,
            migration_bins=energy_migration_bins,
        )
        hdus.append(
            create_energy_dispersion_hdu(
                edisp,
                true_energy_bins=true_energy_bins,
                migration_bins=energy_migration_bins,
                fov_offset_bins=fov_offset_bins,
                extname="ENERGY_DISPERSION" + label,
            )
        )

    bias_resolution = energy_bias_resolution(
        gammas[gammas["selected"]], reco_energy_bins, energy_type="reco"
    )
    ang_res = angular_resolution(gammas[gammas["selected_gh"]],
                                 reco_energy_bins,
                                 energy_type="reco")
    psf = psf_table(
        gammas[gammas["selected_gh"]],
        true_energy_bins,
        fov_offset_bins=fov_offset_bins,
        source_offset_bins=source_offset_bins,
    )

    background_rate = background_2d(
        background[background['selected_gh']],
        reco_energy_bins,
        fov_offset_bins=np.arange(0, 11) * u.deg,
        t_obs=T_OBS,
    )

    hdus.append(create_background_2d_hdu(
        background_rate,
        reco_energy_bins,
        fov_offset_bins=np.arange(0, 11) * u.deg,
    ))
    hdus.append(create_psf_table_hdu(
        psf, true_energy_bins, source_offset_bins, fov_offset_bins,
    ))
    hdus.append(create_rad_max_hdu(
        theta_cuts["cut"][:, np.newaxis], theta_bins, fov_offset_bins
    ))
    hdus.append(fits.BinTableHDU(ang_res, name="ANGULAR_RESOLUTION"))
    hdus.append(fits.BinTableHDU(bias_resolution, name="ENERGY_BIAS_RESOLUTION"))

    log.info('Writing outputfile')
    fits.HDUList(hdus).writeto("pyirf_eventdisplay.fits.gz", overwrite=True)


if __name__ == "__main__":
    main()



import astropy.units as u
from astropy.table import QTable
import numpy as np


def test_psf():
    from pyirf.irf import psf_table
    from pyirf.utils import cone_solid_angle

    np.random.seed(0)

    N = 1000

    TRUE_SIGMA_1 = 0.2
    TRUE_SIGMA_2 = 0.1
    TRUE_SIGMA = np.append(np.full(N, TRUE_SIGMA_1), np.full(N, TRUE_SIGMA_2))

    # toy event data set with just two energies
    # and a psf per energy bin, point-like
    events = QTable(
        {
            "true_energy": np.append(np.full(N, 1), np.full(N, 2)) * u.TeV,
            "true_source_fov_offset": np.zeros(2 * N) * u.deg,
            "theta": np.random.normal(0, TRUE_SIGMA) * u.deg,
        }
    )

    energy_bins = [0, 1.5, 3] * u.TeV
    fov_bins = [0, 1] * u.deg
    source_bins = np.linspace(0, 1, 201) * u.deg

    # We return a table with one row as needed for gadf
    psf = psf_table(events, energy_bins, source_bins, fov_bins)

    # 2 energy bins, 1 fov bin, 200 source distance bins
    assert psf.shape == (2, 1, 200)
    assert psf.unit == u.Unit("sr-1")

    # check that psf is normalized
    bin_solid_angle = np.diff(cone_solid_angle(source_bins))
    assert np.allclose(np.sum(psf * bin_solid_angle, axis=2), 1.0)

    cumulated = np.cumsum(psf * bin_solid_angle, axis=2)

    # first energy and only fov bin
    bin_centers = 0.5 * (source_bins[1:] + source_bins[:-1])
    assert u.isclose(
        bin_centers[np.where(cumulated[0, 0, :] >= 0.68)[0][0]],
        TRUE_SIGMA_1 * u.deg,
        rtol=0.1,
    )

    # second energy and only fov bin
    assert u.isclose(
        bin_centers[np.where(cumulated[1, 0, :] >= 0.68)[0][0]],
        TRUE_SIGMA_2 * u.deg,
        rtol=0.1,
    )



import numpy as np
import astropy.units as u
import pytest


def test_integrate_energy():
    from pyirf.simulations import SimulatedEventsInfo

    info = SimulatedEventsInfo(
        n_showers=int(1e6),
        energy_min=100 * u.GeV,
        energy_max=10 * u.TeV,
        max_impact=500 * u.m,
        spectral_index=-2,
        viewcone=10 * u.deg,
    )
    # simplest case, no bins  outside e_min, e_max
    energy_bins = np.geomspace(info.energy_min, info.energy_max, 20)
    assert np.isclose(
        info.calculate_n_showers_per_energy(energy_bins).sum(),
        info.n_showers,
    )

    # simple case, e_min and e_max on bin edges
    energy_bins = np.geomspace(info.energy_min, info.energy_max, 20)
    energy_bins = np.append(0.9 * info.energy_min, energy_bins)
    energy_bins = np.append(energy_bins, 1.1 * info.energy_max)

    events = info.calculate_n_showers_per_energy(energy_bins)
    assert np.isclose(info.n_showers, events.sum())

    assert events[0] == 0
    assert events[-1] == 0

    # complex case, e_min and e_max inside bins
    energy_bins = np.geomspace(
        0.5 * info.energy_min, 2 * info.energy_max, 5
    )
    events = info.calculate_n_showers_per_energy(energy_bins)
    assert np.isclose(info.n_showers, events.sum())
    assert events[0] > 0
    assert events[-1] > 0


def test_integrate_energy_fov():
    from pyirf.simulations import SimulatedEventsInfo

    # simple case, max viewcone on bin edge
    info = SimulatedEventsInfo(
        n_showers=int(1e6),
        energy_min=100 * u.GeV,
        energy_max=10 * u.TeV,
        max_impact=500 * u.m,
        spectral_index=-2,
        viewcone=10 * u.deg,
    )

    fov_bins = [0, 10, 20] * u.deg
    energy_bins = np.geomspace(info.energy_min, info.energy_max, 20)

    n_events = info.calculate_n_showers_per_energy_and_fov(energy_bins, fov_bins)

    assert np.all(n_events[:, 1:] == 0)
    assert np.isclose(np.sum(n_events), int(1e6))

    # viewcone inside of bin
    info = SimulatedEventsInfo(
        n_showers=int(1e6),
        energy_min=100 * u.GeV,
        energy_max=10 * u.TeV,
        max_impact=500 * u.m,
        spectral_index=-2,
        viewcone=10 * u.deg,
    )

    fov_bins = [0, 9, 11, 20] * u.deg
    energy_bins = np.geomspace(info.energy_min, info.energy_max, 20)
    n_events = info.calculate_n_showers_per_energy_and_fov(energy_bins, fov_bins)

    assert np.all(n_events[:, 1:2] > 0)
    assert np.all(n_events[:, 2:] == 0)
    assert np.isclose(np.sum(n_events), int(1e6))


def test_integrate_energy_fov_pointlike():
    from pyirf.simulations import SimulatedEventsInfo

    info = SimulatedEventsInfo(
        n_showers=int(1e6),
        energy_min=100 * u.GeV,
        energy_max=10 * u.TeV,
        max_impact=500 * u.m,
        spectral_index=-2,
        viewcone=0 * u.deg,
    )

    fov_bins = [0, 9, 11, 20] * u.deg
    energy_bins = np.geomspace(info.energy_min, info.energy_max, 20)

    # make sure we raise an error on invalid input
    with pytest.raises(ValueError):
        info.calculate_n_showers_per_energy_and_fov(energy_bins, fov_bins)



from .energy_bias_resolution import energy_bias_resolution, energy_bias_resolution_from_energy_dispersion
from .angular_resolution import angular_resolution


__all__ = [
    "energy_bias_resolution",
    "energy_bias_resolution_from_energy_dispersion",
    "angular_resolution",
]



try:
    import gammapy
except ImportError:
    raise ImportError('You need gammapy installed to use this module of pyirf') from None

from gammapy.irf import EffectiveAreaTable2D, PSF3D, EnergyDispersion2D
from gammapy.maps import MapAxis
import astropy.units as u



def _create_offset_axis(fov_offset_bins):
    return MapAxis.from_edges(fov_offset_bins, name="offset")

def _create_energy_axis_true(true_energy_bins):
    return MapAxis.from_edges(true_energy_bins, name="energy_true")


@u.quantity_input(
    effective_area=u.m ** 2, true_energy_bins=u.TeV, fov_offset_bins=u.deg
)
def create_effective_area_table_2d(
    effective_area,
    true_energy_bins,
    fov_offset_bins,
):
    '''
    Create a :py:class:`gammapy.irf.EffectiveAreaTable2D` from pyirf outputs.

    Parameters
    ----------
    effective_area: astropy.units.Quantity[area]
        Effective area array, must have shape (n_energy_bins, n_fov_offset_bins)
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.

    Returns
    -------
    gammapy.irf.EffectiveAreaTable2D
    aeff2d: gammapy.irf.EffectiveAreaTable2D
    '''
    offset_axis = _create_offset_axis(fov_offset_bins)
    energy_axis_true = _create_energy_axis_true(true_energy_bins)

    return EffectiveAreaTable2D(
        axes = [energy_axis_true,
                offset_axis],
        data=effective_area,
    )



@u.quantity_input(
    psf=u.sr ** -1,
    true_energy_bins=u.TeV,
    source_offset_bins=u.deg,
    fov_offset_bins=u.deg,
)
def create_psf_3d(
    psf,
    true_energy_bins,
    source_offset_bins,
    fov_offset_bins,
):
    """
    Create a :py:class:`gammapy.irf.PSF3D` from pyirf outputs.

    Parameters
    ----------
    psf: astropy.units.Quantity[(solid angle)^-1]
        Point spread function array, must have shape
        (n_energy_bins, n_fov_offset_bins, n_source_offset_bins)
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    source_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the source offset.
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.

    Returns
    -------
    psf: gammapy.irf.PSF3D
    """
    offset_axis = _create_offset_axis(fov_offset_bins)
    energy_axis_true = _create_energy_axis_true(true_energy_bins)
    rad_axis = MapAxis.from_edges(source_offset_bins, name='rad')

    return PSF3D(
        axes = [energy_axis_true,
                offset_axis,
                rad_axis],
        data = psf
    )


@u.quantity_input(
    true_energy_bins=u.TeV, fov_offset_bins=u.deg,
)
def create_energy_dispersion_2d(
    energy_dispersion,
    true_energy_bins,
    migration_bins,
    fov_offset_bins,
):
    """
    Create a :py:class:`gammapy.irf.EnergyDispersion2D` from pyirf outputs.

    Parameters
    ----------
    energy_dispersion: numpy.ndarray
        Energy dispersion array, must have shape
        (n_energy_bins, n_migra_bins, n_source_offset_bins)
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    migration_bins: numpy.ndarray
        Bin edges for the relative energy migration (``reco_energy / true_energy``)
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.

    Returns
    -------
    edisp: gammapy.irf.EnergyDispersion2D
    """
    offset_axis = _create_offset_axis(fov_offset_bins)
    energy_axis_true = _create_energy_axis_true(true_energy_bins)
    migra_axis = MapAxis.from_edges(migration_bins, name="migra")

    return EnergyDispersion2D(
        axes = [energy_axis_true,
                migra_axis,
                offset_axis],
        data = energy_dispersion,
    )



"""Tests for base interpolator classes"""
import numpy as np


def test_GridDataInterpolator_1DGrid():
    """Test checks performed to enforce correct data structure"""
    from pyirf.interpolation import GridDataInterpolator

    grid_points = np.array([[0], [1]])
    target_point = np.array([[0.5]])

    dummy_data1 = np.array([[[0, 1], [1, 1]], [[0, 2], [2, 3]], [[0, 3], [3, 5]]])
    dummy_data2 = np.array([[[0, 2], [2, 2]], [[0, 4], [4, 6]], [[0, 6], [6, 10]]])

    dummy_data = np.array([dummy_data1, dummy_data2])

    Interpolator = GridDataInterpolator(
        grid_points=grid_points,
        params=dummy_data,
    )
    interpolant = Interpolator(target_point)

    dummy_data_target = 1.5 * dummy_data1

    assert np.allclose(interpolant, dummy_data_target)
    assert interpolant.shape == (1, *dummy_data.shape[1:])


def test_GridDataInterpolator_2DGrid():
    """Test checks performed to enforce correct data structure"""
    from pyirf.interpolation import GridDataInterpolator

    grid_points = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
    target_point = np.array([[0.5, 0.5]])

    dummy_data1 = np.array([[[0, 1], [1, 1]], [[0, 2], [2, 3]], [[0, 3], [3, 5]]])
    dummy_data2 = np.array([[[0, 2], [1, 2]], [[0, 4], [2, 6]], [[0, 6], [3, 10]]])
    dummy_data3 = np.array([[[0, 1], [2, 1]], [[0, 2], [4, 3]], [[0, 3], [6, 5]]])
    dummy_data4 = np.array([[[0, 2], [2, 2]], [[0, 4], [4, 6]], [[0, 6], [6, 10]]])

    dummy_data = np.array([dummy_data1, dummy_data2, dummy_data3, dummy_data4])

    Interpolator = GridDataInterpolator(
        grid_points=grid_points,
        params=dummy_data,
    )
    interpolant = Interpolator(target_point)

    dummy_data_target = 1.5 * dummy_data1

    assert np.allclose(interpolant, dummy_data_target)
    assert interpolant.shape == (1, *dummy_data.shape[1:])



import numpy as np
import astropy.units as u

from ..utils import cone_solid_angle


def psf_table(events, true_energy_bins, source_offset_bins, fov_offset_bins):
    """
    Calculate the table based PSF (radially symmetrical bins around the true source)
    """

    array = np.column_stack(
        [
            events["true_energy"].to_value(u.TeV),
            events["true_source_fov_offset"].to_value(u.deg),
            events["theta"].to_value(u.deg),
        ]
    )

    hist, _ = np.histogramdd(
        array,
        [
            true_energy_bins.to_value(u.TeV),
            fov_offset_bins.to_value(u.deg),
            source_offset_bins.to_value(u.deg),
        ],
    )

    psf = _normalize_psf(hist, source_offset_bins)
    return psf


def _normalize_psf(hist, source_offset_bins):
    """Normalize the psf histogram to a probability densitity over solid angle"""
    solid_angle = np.diff(cone_solid_angle(source_offset_bins))

    # ignore numpy zero division warning
    with np.errstate(invalid="ignore"):

        # normalize over the theta axis
        n_events = hist.sum(axis=2)
        # normalize and replace nans with 0
        psf = np.nan_to_num(hist / n_events[:, :, np.newaxis])

    return psf / solid_angle



import numpy as np
from scipy.stats import norm
from astropy.table import QTable
import astropy.units as u

from ..binning import calculate_bin_indices, UNDERFLOW_INDEX, OVERFLOW_INDEX


NORM_LOWER_SIGMA, NORM_UPPER_SIGMA = norm(0, 1).cdf([-1, 1])
ONE_SIGMA_COVERAGE = NORM_UPPER_SIGMA - NORM_LOWER_SIGMA
MEDIAN = 0.5


def energy_resolution_absolute_68(rel_error):
    """Calculate the energy resolution as the central 68% interval.

    Utility function for pyirf.benchmarks.energy_bias_resolution

    Parameters
    ----------
    rel_error : numpy.ndarray(dtype=float, ndim=1)
        Array of float on which the quantile is calculated.

    Returns
    -------
    resolution: numpy.ndarray(dtype=float, ndim=1)
        Array containing the 68% intervals
    """
    return np.nanquantile(np.abs(rel_error), ONE_SIGMA_COVERAGE)


def inter_quantile_distance(rel_error):
    """Calculate the energy resolution as the half of the 68% containment.

    Percentile equivalent of the standard deviation.
    Utility function for pyirf.benchmarks.energy_bias_resolution

    Parameters
    ----------
    rel_error : numpy.ndarray(dtype=float, ndim=1)
        Array of float on which the quantile is calculated.

    Returns
    -------
    resolution: numpy.ndarray(dtype=float, ndim=1)
        Array containing the resolution values.
    """
    upper_sigma = np.nanquantile(rel_error, NORM_UPPER_SIGMA)
    lower_sigma = np.nanquantile(rel_error, NORM_LOWER_SIGMA)
    resolution = 0.5 * (upper_sigma - lower_sigma)
    return resolution


def energy_bias_resolution(
    events,
    energy_bins,
    energy_type="true",
    bias_function=np.median,
    resolution_function=inter_quantile_distance,
):
    """
    Calculate bias and energy resolution.

    Parameters
    ----------
    events: astropy.table.QTable
        Astropy Table object containing the reconstructed events information.
    energy_bins: numpy.ndarray(dtype=float, ndim=1)
        Bin edges in energy.
    energy_type: str
        Either "true" or "reco" energy.
        Default is "true".
    bias_function: callable
        Function used to calculate the energy bias
    resolution_function: callable
        Function used to calculate the energy resolution

    Returns
    -------
    result : astropy.table.QTable
        QTable containing the energy bias and resolution
        per each bin in true energy.
    """

    # create a table to make use of groupby operations
    table = QTable(events[["true_energy", "reco_energy"]], copy=False)
    table["rel_error"] = (events["reco_energy"] / events["true_energy"]).to_value(u.one) - 1

    energy_key = f"{energy_type}_energy"

    result = QTable()
    result[f"{energy_key}_low"] = energy_bins[:-1]
    result[f"{energy_key}_high"] = energy_bins[1:]
    result[f"{energy_key}_center"] = 0.5 * (energy_bins[:-1] + energy_bins[1:])

    result["n_events"] = 0
    result["bias"] = np.nan
    result["resolution"] = np.nan

    if not len(events):
        # if we get an empty input (no selected events available)
        # we return the table filled with NaNs
        return result


    # use groupby operations to calculate the percentile in each bin
    bin_index, valid = calculate_bin_indices(table[energy_key], energy_bins)
    by_bin = table.group_by(bin_index)

    # use groupby operations to calculate the percentile in each bin
    by_bin = table[valid].group_by(bin_index[valid])
    for bin_idx, group in zip(by_bin.groups.keys, by_bin.groups):
        result["n_events"][bin_idx] = len(group)
        result["bias"][bin_idx] = bias_function(group["rel_error"])
        result["resolution"][bin_idx] = resolution_function(group["rel_error"])
    return result

def energy_bias_resolution_from_energy_dispersion(
    energy_dispersion,
    migration_bins,
):
    """
    Calculate bias and energy resolution.

    Parameters
    ----------
    edisp:
        Energy dispersion matrix of shape
        (n_energy_bins, n_migra_bins, n_source_offset_bins)
    migration_bins: numpy.ndarray
        Bin edges for the relative energy migration (``reco_energy / true_energy``)
    """

    cdf = np.cumsum(energy_dispersion, axis=1)

    n_energy_bins, _, n_fov_bins = energy_dispersion.shape

    bias = np.zeros((n_energy_bins, n_fov_bins))
    resolution = np.zeros((n_energy_bins, n_fov_bins))

    for energy_bin in range(n_energy_bins):
        for fov_bin in range(n_fov_bins):

            low, median, high = np.interp(
                [NORM_LOWER_SIGMA, MEDIAN, NORM_UPPER_SIGMA],
                cdf[energy_bin, :, fov_bin],
                migration_bins[1:] # cdf is defined at upper bin edge
            )
            bias[energy_bin, fov_bin] = median - 1
            resolution[energy_bin, fov_bin] = 0.5 * (high - low)

    return bias, resolution



import warnings
import numpy as np
import astropy.units as u
from ..binning import resample_histogram1d


__all__ = [
    "energy_dispersion",
    "energy_dispersion_to_migration"
]


def _normalize_hist(hist):
    # make sure we do not mutate the input array
    hist = hist.copy()

    # calculate number of events along the N_MIGRA axis to get events
    # per energy per fov
    norm = hist.sum(axis=1)

    with np.errstate(invalid="ignore"):
        # hist shape is (N_E, N_MIGRA, N_FOV), norm shape is (N_E, N_FOV)
        # so we need to add a new axis in the middle to get (N_E, 1, N_FOV)
        # for broadcasting
        hist = hist / norm[:, np.newaxis, :]

    return np.nan_to_num(hist)


def energy_dispersion(
    selected_events, true_energy_bins, fov_offset_bins, migration_bins,
):
    """
    Calculate energy dispersion for the given DL2 event list.
    Energy dispersion is defined as the probability of finding an event
    at a given relative deviation ``(reco_energy / true_energy)`` for a given
    true energy.

    Parameters
    ----------
    selected_events: astropy.table.QTable
        Table of the DL2 events.
        Required columns: ``reco_energy``, ``true_energy``, ``true_source_fov_offset``.
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    migration_bins: astropy.units.Quantity[energy]
        Bin edges in relative deviation, recommended range: [0.2, 5]
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.

    Returns
    -------
    energy_dispersion: numpy.ndarray
        Energy dispersion matrix
        with shape (n_true_energy_bins, n_migration_bins, n_fov_ofset_bins)
    """
    mu = (selected_events["reco_energy"] / selected_events["true_energy"]).to_value(
        u.one
    )

    energy_dispersion, _ = np.histogramdd(
        np.column_stack(
            [
                selected_events["true_energy"].to_value(u.TeV),
                mu,
                selected_events["true_source_fov_offset"].to_value(u.deg),
            ]
        ),
        bins=[
            true_energy_bins.to_value(u.TeV),
            migration_bins,
            fov_offset_bins.to_value(u.deg),
        ],
    )

    n_events_per_energy = energy_dispersion.sum(axis=1)
    energy_dispersion = _normalize_hist(energy_dispersion)

    return energy_dispersion


def energy_dispersion_to_migration(
    dispersion_matrix,
    disp_true_energy_edges,
    disp_migration_edges,
    new_true_energy_edges,
    new_reco_energy_edges,
):
    """
    Construct a energy migration matrix from a energy
    dispersion matrix.
    Depending on the new energy ranges, the sum over the first axis
    can be smaller than 1.
    The new true energy bins need to be a subset of the old range,
    extrapolation is not supported.
    New reconstruction bins outside of the old migration range are filled with
    zeros.

    Parameters
    ----------
    dispersion_matrix: numpy.ndarray
        Energy dispersion_matrix
    disp_true_energy_edges: astropy.units.Quantity[energy]
        True energy edges matching the first dimension of the dispersion matrix
    disp_migration_edges: numpy.ndarray
        Migration edges matching the second dimension of the dispersion matrix
    new_true_energy_edges: astropy.units.Quantity[energy]
        True energy edges matching the first dimension of the output
    new_reco_energy_edges: astropy.units.Quantity[energy]
        Reco energy edges matching the second dimension of the output

    Returns:
    --------
    migration_matrix: numpy.ndarray
        Three-dimensional energy migration matrix. The third dimension
        equals the fov offset dimension of the energy dispersion matrix.
    """

    migration_matrix = np.zeros((
        len(new_true_energy_edges) - 1,
        len(new_reco_energy_edges) - 1,
        dispersion_matrix.shape[2],
    ))

    true_energy_interpolation = resample_histogram1d(
        dispersion_matrix,
        disp_true_energy_edges,
        new_true_energy_edges,
        axis=0,
    )

    norm = np.sum(true_energy_interpolation, axis=1, keepdims=True)
    norm[norm == 0] = 1
    true_energy_interpolation /= norm

    for idx, e_true in enumerate(
        (new_true_energy_edges[1:] + new_true_energy_edges[:-1]) / 2
    ):

        # get migration for the new true energy bin
        e_true_dispersion = true_energy_interpolation[idx]

        with warnings.catch_warnings():
            # silence inf/inf division warning
            warnings.filterwarnings('ignore', 'invalid value encountered in true_divide')
            interpolation_edges = new_reco_energy_edges / e_true

        y = resample_histogram1d(
            e_true_dispersion,
            disp_migration_edges,
            interpolation_edges,
            axis=0,
        )

        migration_matrix[idx, :, :] = y

    return migration_matrix



import pathlib
import re

import pytest
from astropy.units import Quantity

from gammapy.irf import load_irf_dict_from_file

PROD5_IRF_PATH = pathlib.Path(__file__).parent.parent / "irfs/"


@pytest.fixture(scope="session")
def prod5_irfs():
    if not PROD5_IRF_PATH.exists():
        pytest.fail(
            "Test IRF files missing, you need to download them using "
            "`python download_irfs.py` in pyirfs root directory."
        )

    # Get dict of {ZEN_PNT, IRF} pairs for each file in ./irfs
    irfs = {
        Quantity(re.search(r"\d{2}deg", str(irf_file)).group()): load_irf_dict_from_file(
            irf_file
        )
        for irf_file in PROD5_IRF_PATH.glob("*.fits.gz")
    }

    assert len(irfs) == 3
    for key in ["aeff", "psf", "edisp"]:
        for irf in irfs.values():
            assert key in irf

    # Sort dict by zenith angle
    return dict(sorted(irfs.items()))



from astropy.table import QTable
import astropy.units as u
import numpy as np
from scipy.stats import norm
from itertools import product


def test_empty_bias_resolution():
    from pyirf.benchmarks import energy_bias_resolution

    events = QTable({
        'true_energy': [] * u.TeV,
        'reco_energy': [] * u.TeV,
    })

    table = energy_bias_resolution(
        events,
        [1, 10, 100] * u.TeV
    )

    assert np.all(np.isnan(table["bias"]))
    assert np.all(np.isnan(table["resolution"]))


def test_absolute_68():
    from pyirf.benchmarks.energy_bias_resolution import energy_resolution_absolute_68
    rng = np.random.default_rng(0)
    values = rng.normal(0, 1, 1000)
    assert np.isclose(energy_resolution_absolute_68(values), 1, rtol=0.01)


def test_energy_bias_resolution():
    from pyirf.benchmarks import energy_bias_resolution

    np.random.seed(1337)

    TRUE_RES_1 = 0.2
    TRUE_RES_2 = 0.05
    TRUE_BIAS_1 = 0.1
    TRUE_BIAS_2 = -0.05

    true_bias = np.append(np.full(1000, TRUE_BIAS_1), np.full(1000, TRUE_BIAS_2))
    true_resolution = np.append(np.full(1000, TRUE_RES_1), np.full(1000, TRUE_RES_2))

    true_energy = np.concatenate([
        [0.5], # below bin 1 to test with underflow
        np.full(999, 5.0),
        np.full(999, 50.0),
        [500], # above bin 2 to test with overflow
    ]) * u.TeV
    reco_energy = true_energy * (1 + np.random.normal(true_bias, true_resolution))

    events = QTable({
        'true_energy': true_energy,
        'reco_energy': reco_energy,
    })

    bias_resolution = energy_bias_resolution(
        events,
        [1, 10, 100] * u.TeV
    )

    bias = bias_resolution['bias'].quantity
    resolution = bias_resolution['resolution'].quantity

    assert len(bias) == len(resolution) == 2

    assert u.isclose(bias[0], TRUE_BIAS_1, rtol=0.05)
    assert u.isclose(bias[1], TRUE_BIAS_2, rtol=0.05)
    assert u.isclose(resolution[0], TRUE_RES_1, rtol=0.05)
    assert u.isclose(resolution[1], TRUE_RES_2, rtol=0.05)



def test_energy_bias_resolution():
    from pyirf.benchmarks import energy_bias_resolution_from_energy_dispersion
    from pyirf.binning import bin_center

    # create a toy energy dispersion
    n_migra_bins = 500
    true_bias = np.array([
        [0.5, 0.6],
        [0, 0.1],
        [-0.1, -0.2],
    ])
    true_resolution = np.array([
        [0.4, 0.5],
        [0.2, 0.3],
        [0.1, 0.15],
    ])

    n_energy_bins, n_fov_bins = true_bias.shape
    energy_bins = np.geomspace(10, 1000, n_energy_bins + 1)
    energy_center = bin_center(energy_bins)
    migra_bins = np.geomspace(0.2, 5, n_migra_bins + 1)

    cdf = np.empty((n_energy_bins, n_migra_bins + 1, n_fov_bins))
    for energy_bin, fov_bin in product(range(n_energy_bins), range(n_fov_bins)):

        energy = energy_center[energy_bin]
        mu = (1 + true_bias[energy_bin, fov_bin]) * energy
        sigma = true_resolution[energy_bin, fov_bin] * energy
        reco_energy = migra_bins * energy

        cdf[energy_bin, :, fov_bin] = norm.cdf(reco_energy, mu, sigma)

    edisp = cdf[:, 1:, :] - cdf[:, :-1, :]

    bias, resolution = energy_bias_resolution_from_energy_dispersion(
        edisp,
        migra_bins,
    )

    assert np.allclose(bias, true_bias, atol=0.01)
    assert np.allclose(resolution, true_resolution, atol=0.01)



class IRFException(Exception):
    pass


class MissingColumns(IRFException):
    """
    Exception to be raised when a table is missing a required column.
    """

    def __init__(self, columns):
        super().__init__(f"Table is missing required columns {columns}")


class WrongColumnUnit(IRFException):
    """
    Exception to be raised when a column of a table has the wrong unit.
    """

    def __init__(self, column, unit, expected):
        super().__init__(
            f'Unit {unit} of column "{column}"'
            f' has incompatible unit "{unit}", expected {expected}'
            f" required column {column}"
        )



import pytest
import numpy as np


def test_lima():
    from pyirf.statistics import li_ma_significance

    assert li_ma_significance(10, 2, 0.2) > 5
    assert li_ma_significance(10, 0, 0.2) > 5
    assert li_ma_significance(1, 6, 0.2) == 0


def test_lima_gammapy():
    pytest.importorskip("gammapy")
    from gammapy.stats import WStatCountsStatistic
    from pyirf.statistics import li_ma_significance

    n_ons = [100, 50, 10]
    n_offs = [10, 20, 30]
    alphas = [2, 1, 0.2]
    for n_on, n_off, alpha in zip(n_ons, n_offs, alphas):
        sig_gammapy = WStatCountsStatistic(n_on, n_off, alpha).sqrt_ts
        assert np.isclose(li_ma_significance(n_on, n_off, alpha), sig_gammapy)


def test_lima_accuracy():
    from pyirf.statistics import li_ma_significance

    noff = 1e7
    nexcess = 1e4

    res_f64 = li_ma_significance(
        np.float64(noff + nexcess), np.float64(noff / 0.2), 0.2
    )
    res_f32 = li_ma_significance(
        np.float32(noff + nexcess), np.float32(noff / 0.2), 0.2
    )

    assert np.isclose(res_f64, res_f32)



import numpy as np
from scipy.interpolate import griddata, interp1d

from .base_interpolators import BinnedInterpolator

__all__ = ["QuantileInterpolator"]


def cdf_values(bin_contents):
    """
    compute cdf values and assure they are normed to unity
    """
    cdfs = np.cumsum(bin_contents, axis=-1)

    # assure the last cdf value is 1, ignore errors for empty pdfs as they are reset to 0 by nan_to_num
    with np.errstate(invalid="ignore"):
        cdfs = np.nan_to_num(cdfs / np.max(cdfs, axis=-1)[..., np.newaxis])

    return cdfs


def ppf_values(bin_mids, cdfs, quantiles):
    """
    Compute ppfs from cdfs and interpolate them to the desired interpolation point

    Parameters
    ----------
    bin_mids: numpy.ndarray, shape=(M)
        Bin-mids for each bin along interpolation axis

    cdfs: numpy.ndarray, shape=(N,...,M)
        Corresponding cdf-values for all quantiles

    quantiles: numpy.ndarray, shape=(L)
        Quantiles for which the ppf-values should be estimated

    Returns
    -------
    ppfs: numpy.ndarray, shape=(1,...,L)
        Corresponding ppf-values for all quantiles at the target interpolation point
    """

    def cropped_interp(cdf):
        """
        create ppf-values through inverse interpolation of the cdf, avoiding repeating 0 and 1 entries
        around the first and last bins as well as repeating values due to zero bins
        in between filled bins. Both cases would result in division by zero errors when computing
        the interpolation polynom.
        """

        # Find last 0 and first 1 entry
        last_0 = np.nonzero(cdf == 0)[0][-1] if cdf[0] == 0 else 0
        first_1 = np.nonzero(cdf == 1.0)[0][0]

        # Predefine selection mask
        selection = np.ones(len(cdf), dtype=bool)

        # Keep only the first of subsequently matching values
        selection[1:] = cdf[1:] != cdf[:-1]

        # Keep only the last 0 and first 1 entry
        selection[:last_0] = False
        selection[last_0] = True
        selection[first_1 + 1 :] = False
        selection[first_1] = True

        # create ppf values from selected bins
        return interp1d(
            cdf[selection],
            bin_mids[selection],
            bounds_error=False,
            fill_value="extrapolate",
        )(quantiles)

    # create ppf values from cdf samples via interpolation of the cdfs
    # return nan for empty pdfs
    ppfs = np.apply_along_axis(
        lambda cdf: cropped_interp(cdf)
        if np.sum(cdf) > 0
        else np.full_like(quantiles, np.nan),
        -1,
        cdfs,
    )
    # nD interpolation of ppf values
    return ppfs


def pdf_from_ppf(bin_edges, interp_ppfs, quantiles):
    """
    Reconstruct pdf from ppf and evaluate at desired points.

    Parameters
    ----------
    bin_edges: numpy.ndarray, shape=(M+1)
        Edges of the bins in which the final pdf should be binned

    interp_ppfs: numpy.ndarray, shape=(1,...,L)
        Corresponding ppf-values for all self.quantiles at the target_point,
        not to be confused with QunatileInterpolators self.ppfs, the ppfs
        computed from the input distributions.

    quantiles: numpy.ndarray, shape=(L)
        Quantiles corresponding to the ppf-values in interp_ppfs

    Returns
    -------
    pdf_values: numpy.ndarray, shape=(1,...,M)
        Recomputed, binned pdf at target_point
    """
    # recalculate pdf values through numerical differentiation
    pdf_interpolant = np.nan_to_num(np.diff(quantiles) / np.diff(interp_ppfs, axis=-1))

    # Unconventional solution to make this usable with np.apply_along_axis for readability
    # The ppf bin-mids are computed since the pdf-values are derived through derivation
    # from the ppf-values
    xyconcat = np.concatenate(
        (interp_ppfs[..., :-1] + np.diff(interp_ppfs) / 2, pdf_interpolant), axis=-1
    )

    def interpolate_ppf(xy):
        ppf = xy[: len(xy) // 2]
        pdf = xy[len(xy) // 2 :]
        interpolate = interp1d(ppf, pdf, bounds_error=False, fill_value=(0, 0))
        result = np.nan_to_num(interpolate(bin_edges[:-1]))
        return np.diff(bin_edges) * result

    # Interpolate pdf samples and evaluate at bin edges, weight with the bin_width to estimate
    # correct bin height via the midpoint rule formulation of the trapezoidal rule
    pdf_values = np.apply_along_axis(interpolate_ppf, -1, xyconcat)

    return pdf_values


def norm_pdf(pdf_values):
    """
    Normalize binned_pdf to a sum of 1
    """
    norm = np.sum(pdf_values, axis=-1)

    # Norm all binned_pdfs to unity that are not empty
    normed_pdf_values = np.divide(
        pdf_values,
        norm[..., np.newaxis],
        out=np.zeros_like(pdf_values),
        where=norm[..., np.newaxis] != 0,
    )

    return normed_pdf_values


class QuantileInterpolator(BinnedInterpolator):
    def __init__(
        self, grid_points, bin_edges, bin_contents, axis, quantile_resolution=1e-3
    ):
        """BinnedInterpolator constructor

        Parameters
        ----------
        grid_points: np.ndarray
            Grid points at which interpolation templates exist
        bin_edges: np.ndarray
            Edges of the data binning
        bin_contents: np.ndarray
            Content of each bin in bin_edges for
            each point in grid_points. First dimesion has to correspond to number
            of grid_points, the dimension indicated by axis has to correspond to number
            of bins for the quantity that should be interpolated
            (e.g. the Migra axis for EDisp)
        axis: int
            Interpolation axis
        quantile_resolution: float
            Spacing between quantiles

        Raises
        ------
        TypeError:
            When bin_edges is not a np.ndarray
        TypeError:
            When bin_content is not a np.ndarray
        ValueError:
            When number of bins in bin_edges and contents bin_contents is
            not matching
        ValueError:
            When number of histograms in bin_contents and points in grid_points
            is not matching

        Note
        ----
            Also calls pyirf.interpolation.BaseInterpolators.__call__
        """
        self.axis = axis

        # Remember input shape
        self.input_shape = bin_contents.shape

        # To have the needed axis always at the last index, as the number of indices is
        # not safely propageted through the interpolation but the last element remains the last element
        bin_contents = np.swapaxes(bin_contents, self.axis, -1)

        # include 0-bin at first position in each pdf to avoid edge-effects where the CDF would otherwise
        # start at a value != 0, also extend edges with one bin to the left
        fill_zeros = np.zeros(shape=bin_contents.shape[:-1])[..., np.newaxis]
        bin_contents = np.concatenate((fill_zeros, bin_contents), axis=-1)
        bin_edges = np.append(bin_edges[0] - np.diff(bin_edges)[0], bin_edges)

        # compute quantiles from quantile_resolution
        self.quantiles = np.linspace(
            0, 1, int(np.round(1 / quantile_resolution, decimals=0))
        )

        super().__init__(
            grid_points=grid_points, bin_edges=bin_edges, bin_contents=bin_contents
        )

        # Compute CDF values
        self.cdfs = cdf_values(self.bin_contents)

        # compute ppf values at quantiles, determine quantile step of [1]
        self.ppfs = ppf_values(self.bin_mids, self.cdfs, self.quantiles)

    def interpolate(self, target_point, **kwargs):
        """
        Takes a grid of binned pdfs for a bunch of different parameters
        and interpolates it to given value of those parameters.
        This function provides an adapted version of the quantile interpolation introduced
        in [1].
        Instead of following this method closely, it implements different approaches to the
        steps shown in Fig. 5 of [1].

        Parameters
        ----------
        target_point: numpy.ndarray, shape=(O)
            Value for which the interpolation is performed (target point)

        **kwargs:
            Currently ignored

        Returns
        -------
        f_new: numpy.ndarray, shape=(1,...,M,...)
            Interpolated and binned pdf

        References
        ----------
        .. [1] B. E. Hollister and A. T. Pang (2013). Interpolation of Non-Gaussian Probability Distributions
               for Ensemble Visualization
               https://engineering.ucsc.edu/sites/default/files/technical-reports/UCSC-SOE-13-13.pdf
        """

        # interpolate ppfs to target point, interpolate quantiles step of [1]
        interpolated_ppfs = griddata(self.grid_points, self.ppfs, target_point)

        # compute pdf values for all bins, evaluate interpolant PDF values step of [1], drop the earlier
        # introduced extra bin
        interpolated_pdfs = pdf_from_ppf(
            self.bin_edges, interpolated_ppfs, self.quantiles
        )[..., 1:]

        # Renormalize pdf to sum of 1
        normed_interpolated_pdfs = norm_pdf(interpolated_pdfs)

        # Re-swap axes and set all nans to zero
        return np.swapaxes(
            np.nan_to_num(normed_interpolated_pdfs), self.axis, -1
        ).reshape(1, *self.input_shape[1:])



from .effective_area import (
    effective_area,
    effective_area_per_energy_and_fov,
    effective_area_per_energy,
)
from .energy_dispersion import energy_dispersion
from .psf import psf_table
from .background import background_2d

__all__ = [
    "effective_area",
    "effective_area_per_energy",
    "effective_area_per_energy_and_fov",
    "energy_dispersion",
    "psf_table",
    "background_2d",
]



import astropy.units as u
import numpy as np
from astropy.table import QTable


def test_energy_dispersion():
    from pyirf.irf import energy_dispersion

    np.random.seed(0)

    N = 10000
    TRUE_SIGMA_1 = 0.20
    TRUE_SIGMA_2 = 0.10
    TRUE_SIGMA_3 = 0.05

    selected_events = QTable(
        {
            "reco_energy": np.concatenate(
                [
                    np.random.normal(1.0, TRUE_SIGMA_1, size=N) * 0.5,
                    np.random.normal(1.0, TRUE_SIGMA_2, size=N) * 5,
                    np.random.normal(1.0, TRUE_SIGMA_3, size=N) * 50,
                ]
            )
            * u.TeV,
            "true_energy": np.concatenate(
                [np.full(N, 0.5), np.full(N, 5.0), np.full(N, 50.0)]
            )
            * u.TeV,
            "true_source_fov_offset": np.concatenate(
                [
                    np.full(N // 2, 0.2),
                    np.full(N // 2, 1.5),
                    np.full(N // 2, 0.2),
                    np.full(N // 2, 1.5),
                    np.full(N // 2, 0.2),
                    np.full(N // 2, 1.5),
                ]
            )
            * u.deg,
        }
    )

    true_energy_bins = np.array([0.1, 1.0, 10.0, 100]) * u.TeV
    fov_offset_bins = np.array([0, 1, 2]) * u.deg
    migration_bins = np.linspace(0, 2, 1001)

    result = energy_dispersion(
        selected_events, true_energy_bins, fov_offset_bins, migration_bins
    )

    assert result.shape == (3, 1000, 2)
    assert np.isclose(result.sum(), 6.0)

    cumulative_sum = np.cumsum(result, axis=1)
    bin_centers = 0.5 * (migration_bins[1:] + migration_bins[:-1])
    assert np.isclose(
        TRUE_SIGMA_1,
        (
            bin_centers[np.where(cumulative_sum[0, :, :] >= 0.84)[0][0]]
            - bin_centers[np.where(cumulative_sum[0, :, :] >= 0.16)[0][0]]
        )
        / 2,
        rtol=0.1,
    )
    assert np.isclose(
        TRUE_SIGMA_2,
        (
            bin_centers[np.where(cumulative_sum[1, :, :] >= 0.84)[0][0]]
            - bin_centers[np.where(cumulative_sum[1, :, :] >= 0.16)[0][0]]
        )
        / 2,
        rtol=0.1,
    )
    assert np.isclose(
        TRUE_SIGMA_3,
        (
            bin_centers[np.where(cumulative_sum[2, :, :] >= 0.84)[0][0]]
            - bin_centers[np.where(cumulative_sum[2, :, :] >= 0.16)[0][0]]
        )
        / 2,
        rtol=0.1,
    )


def test_energy_dispersion_to_migration():
    from pyirf.irf import energy_dispersion
    from pyirf.irf.energy_dispersion import energy_dispersion_to_migration

    np.random.seed(0)
    N = 10000
    true_energy_bins = 10**np.arange(np.log10(0.2), np.log10(200), 1/10) * u.TeV

    fov_offset_bins = np.array([0, 1, 2]) * u.deg
    migration_bins = np.linspace(0, 2, 101)

    true_energy = np.random.uniform(
        true_energy_bins[0].value,
        true_energy_bins[-1].value,
        size=N
    ) * u.TeV
    reco_energy = true_energy * np.random.uniform(0.5, 1.5, size=N)

    selected_events = QTable(
        {
            "reco_energy": reco_energy,
            "true_energy": true_energy,
            "true_source_fov_offset": np.concatenate(
                [
                    np.full(N // 2, 0.2),
                    np.full(N // 2, 1.5),
                ]
            )
            * u.deg,
        }
    )

    dispersion_matrix = energy_dispersion(
        selected_events, true_energy_bins, fov_offset_bins, migration_bins
    )

    # migration matrix selecting a limited energy band with different binsizes
    new_true_energy_bins = 10**np.arange(np.log10(2), np.log10(20), 1/5) * u.TeV
    new_reco_energy_bins = 10**np.arange(np.log10(2), np.log10(20), 1/5) * u.TeV
    migration_matrix = energy_dispersion_to_migration(
        dispersion_matrix,
        true_energy_bins,
        migration_bins,
        new_true_energy_bins,
        new_reco_energy_bins,

    )

    # test dimension
    assert migration_matrix.shape[0] == len(new_true_energy_bins) - 1
    assert migration_matrix.shape[1] == len(new_reco_energy_bins) - 1
    assert migration_matrix.shape[2] == dispersion_matrix.shape[2]

    # test that all migrations are included for central energies
    assert np.isclose(migration_matrix.sum(axis=1).max(), 1, rtol=0.01)

    # test that migrations dont always sum to 1 (since some energies are
    # not included in the matrix)
    assert migration_matrix.sum() < (len(new_true_energy_bins) - 1) * (len(fov_offset_bins) - 1)
    assert np.all(np.isfinite(migration_matrix))


def test_overflow_bins_migration_matrix():
    from pyirf.irf import energy_dispersion
    from pyirf.irf.energy_dispersion import energy_dispersion_to_migration
    from pyirf.binning import add_overflow_bins

    np.random.seed(0)
    N = 10000

    # add under/overflow bins
    bins = 10 ** np.arange(
        np.log10(0.2),
        np.log10(200),
        1 / 10,
    ) * u.TeV
    true_energy_bins = add_overflow_bins(bins, positive=False)

    fov_offset_bins = np.array([0, 1, 2]) * u.deg
    migration_bins = np.linspace(0, 2, 101)

    true_energy = np.random.uniform(
        true_energy_bins[1].value,
        true_energy_bins[-2].value,
        size=N
    ) * u.TeV
    reco_energy = true_energy * np.random.uniform(0.5, 1.5, size=N)

    selected_events = QTable(
        {
            "reco_energy": reco_energy,
            "true_energy": true_energy,
            "true_source_fov_offset": np.concatenate(
                [
                    np.full(N // 2, 0.2),
                    np.full(N // 2, 1.5),
                ]
            )
            * u.deg,
        }
    )

    dispersion_matrix = energy_dispersion(
        selected_events, true_energy_bins, fov_offset_bins, migration_bins
    )

    migration_matrix = energy_dispersion_to_migration(
        dispersion_matrix,
        true_energy_bins,
        migration_bins,
        true_energy_bins,
        true_energy_bins,
    )

    # migration into underflow bin is not empty
    assert np.sum(migration_matrix[:, 0, :]) > 0
    # migration from underflow bin is empty
    assert np.sum(migration_matrix[0, :, :]) == 0

    assert np.all(np.isfinite(migration_matrix))



import astropy.units as u
import pytest
import numpy as np


def test_table_interpolation():

    from pyirf.spectral import TableInterpolationSpectrum

    # log log
    energy = [1, 10, 100] * u.TeV
    flux = [50, 5, 0.5] / (u.GeV * u.m**2 * u.sr * u.s)

    spectrum = TableInterpolationSpectrum(energy, flux)
    assert u.isclose(spectrum(5 * u.TeV), 10 / (u.GeV * u.m**2 * u.sr * u.s))


    # lin lin
    energy = [1, 2, 3] * u.TeV
    flux = [10, 8, 6] / (u.GeV * u.m**2 * u.sr * u.s)

    spectrum = TableInterpolationSpectrum(energy, flux, log_energy=False, log_flux=False)
    assert u.isclose(spectrum(1.5 * u.TeV), 9 / (u.GeV * u.m**2 * u.sr * u.s))



def test_powerlaw_integrate_cone_invalid():
    from pyirf.spectral import PowerLaw, POINT_SOURCE_FLUX_UNIT

    point_source = PowerLaw(
        normalization=1 * POINT_SOURCE_FLUX_UNIT,
        index=-2,
        e_ref=500 * u.GeV,
    )
    with pytest.raises(ValueError):
        point_source.integrate_cone(0 * u.deg, 2 * u.deg)


@pytest.mark.parametrize(
    "outer,expected",
    zip((90 * u.deg, 180 * u.deg), (2 * np.pi * u.sr, 4 * np.pi * u.sr))
)
def test_powerlaw_integrate_cone(outer, expected):
    from pyirf.spectral import PowerLaw, POINT_SOURCE_FLUX_UNIT, DIFFUSE_FLUX_UNIT
    diffuse_flux = PowerLaw(
        normalization=1 * DIFFUSE_FLUX_UNIT,
        index=-2,
        e_ref=500 * u.GeV,
    )


    integrated = diffuse_flux.integrate_cone(0 * u.rad, outer)
    assert integrated.normalization.unit.is_equivalent(POINT_SOURCE_FLUX_UNIT)
    assert u.isclose(integrated.normalization, diffuse_flux.normalization * expected)
    assert integrated.index == diffuse_flux.index
    assert integrated.e_ref == diffuse_flux.e_ref


def test_powerlaw():
    from pyirf.spectral import PowerLaw

    with pytest.raises(TypeError):
        PowerLaw(normalization=1e-10, index=-2)

    with pytest.raises(u.UnitsError):
        PowerLaw(normalization=1e-10 / u.TeV, index=-2)

    with pytest.raises(ValueError):
        PowerLaw(normalization=1e-10 / u.TeV / u.m**2 / u.s, index=2)

    # check we get a reasonable unit out of astropy independent of input unit
    unit = u.TeV**-1 * u.m**-2 * u.s**-1
    power_law = PowerLaw(1e-10 * unit, -2.65)
    assert power_law(1 * u.TeV).unit == unit
    assert power_law(1 * u.GeV).unit == unit



import numpy as np
import astropy.units as u
from astropy.table import QTable
import pytest


def test_is_scalar():
    from pyirf.utils import is_scalar

    assert is_scalar(1.0)
    assert is_scalar(5 * u.m)
    assert is_scalar(np.array(5))

    assert not is_scalar([1, 2, 3])
    assert not is_scalar([1, 2, 3] * u.m)
    assert not is_scalar(np.ones(5))
    assert not is_scalar(np.ones((3, 4)))


def test_cone_solid_angle():
    from pyirf.utils import cone_solid_angle

    # whole sphere
    assert u.isclose(cone_solid_angle(np.pi * u.rad), 4 * np.pi * u.sr)

    # half the sphere
    assert u.isclose(cone_solid_angle(90 * u.deg), 2 * np.pi * u.sr)

    # zero
    assert u.isclose(cone_solid_angle(0 * u.deg), 0 * u.sr)


def test_check_table():
    from pyirf.exceptions import MissingColumns, WrongColumnUnit
    from pyirf.utils import check_table

    t = QTable({'bar': [0, 1, 2] * u.TeV})

    with pytest.raises(MissingColumns):
        check_table(t, required_columns=['foo'])

    t = QTable({'bar': [0, 1, 2] * u.TeV})
    with pytest.raises(WrongColumnUnit):
        check_table(t, required_units={'bar': u.m})

    t = QTable({'bar': [0, 1, 2] * u.m})
    with pytest.raises(MissingColumns):
        check_table(t, required_units={'foo': u.m})

    # m is convertible
    check_table(t, required_units={'bar': u.cm})



from astropy.table import QTable
import astropy.units as u
from astropy.io.fits import Header, BinTableHDU
import numpy as np
from astropy.time import Time
import pyirf.binning as binning

from ..version import __version__


__all__ = [
    "create_aeff2d_hdu",
    "create_energy_dispersion_hdu",
    "create_psf_table_hdu",
    "create_rad_max_hdu",
]


DEFAULT_HEADER = Header()
DEFAULT_HEADER["CREATOR"] = f"pyirf v{__version__}"
# fmt: off
DEFAULT_HEADER["HDUDOC"] = "https://github.com/open-gamma-ray-astro/gamma-astro-data-formats"
DEFAULT_HEADER["HDUVERS"] = "0.2"
DEFAULT_HEADER["HDUCLASS"] = "GADF"


def _add_header_cards(header, **header_cards):
    for k, v in header_cards.items():
        header[k] = v


@u.quantity_input(
    effective_area=u.m ** 2, true_energy_bins=u.TeV, fov_offset_bins=u.deg
)
def create_aeff2d_hdu(
    effective_area,
    true_energy_bins,
    fov_offset_bins,
    extname="EFFECTIVE AREA",
    point_like=True,
    **header_cards,
):
    """
    Create a fits binary table HDU in GADF format for effective area.
    See the specification at
    https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/full_enclosure/aeff/index.html

    Parameters
    ----------
    effective_area: astropy.units.Quantity[area]
        Effective area array, must have shape (n_energy_bins, n_fov_offset_bins)
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.
    point_like: bool
        If the provided effective area was calculated after applying a direction cut,
        pass ``True``, else ``False`` for a full-enclosure effective area.
    extname: str
        Name for BinTableHDU
    **header_cards
        Additional metadata to add to the header, use this to set e.g. TELESCOP or
        INSTRUME.
    """
    aeff = QTable()
    aeff["ENERG_LO"], aeff["ENERG_HI"] = binning.split_bin_lo_hi(true_energy_bins[np.newaxis, :].to(u.TeV))
    aeff["THETA_LO"], aeff["THETA_HI"] = binning.split_bin_lo_hi(fov_offset_bins[np.newaxis, :].to(u.deg))
    # transpose because FITS uses opposite dimension order than numpy
    aeff["EFFAREA"] = effective_area.T[np.newaxis, ...].to(u.m ** 2)

    # required header keywords
    header = DEFAULT_HEADER.copy()
    header["HDUCLAS1"] = "RESPONSE"
    header["HDUCLAS2"] = "EFF_AREA"
    header["HDUCLAS3"] = "POINT-LIKE" if point_like else "FULL-ENCLOSURE"
    header["HDUCLAS4"] = "AEFF_2D"
    header["DATE"] = Time.now().utc.iso
    _add_header_cards(header, **header_cards)

    return BinTableHDU(aeff, header=header, name=extname)


@u.quantity_input(
    psf=u.sr ** -1,
    true_energy_bins=u.TeV,
    source_offset_bins=u.deg,
    fov_offset_bins=u.deg,
)
def create_psf_table_hdu(
    psf,
    true_energy_bins,
    source_offset_bins,
    fov_offset_bins,
    extname="PSF",
    **header_cards,
):
    """
    Create a fits binary table HDU in GADF format for the PSF table.
    See the specification at
    https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/full_enclosure/psf/psf_table/index.html

    Parameters
    ----------
    psf: astropy.units.Quantity[(solid angle)^-1]
        Point spread function array, must have shape
        (n_energy_bins, n_fov_offset_bins, n_source_offset_bins)
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    source_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the source offset.
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.
    extname: str
        Name for BinTableHDU
    **header_cards
        Additional metadata to add to the header, use this to set e.g. TELESCOP or
        INSTRUME.
    """

    psf_ = QTable()
    psf_["ENERG_LO"], psf_["ENERG_HI"] = binning.split_bin_lo_hi(true_energy_bins[np.newaxis, :].to(u.TeV))
    psf_["THETA_LO"], psf_["THETA_HI"] = binning.split_bin_lo_hi(fov_offset_bins[np.newaxis, :].to(u.deg))
    psf_["RAD_LO"], psf_["RAD_HI"] = binning.split_bin_lo_hi(source_offset_bins[np.newaxis, :].to(u.deg))
    # transpose as FITS uses opposite dimension order
    psf_["RPSF"] = psf.T[np.newaxis, ...].to(1 / u.sr)

    # required header keywords
    header = DEFAULT_HEADER.copy()
    header["HDUCLAS1"] = "RESPONSE"
    header["HDUCLAS2"] = "PSF"
    header["HDUCLAS3"] = "FULL-ENCLOSURE"
    header["HDUCLAS4"] = "PSF_TABLE"
    header["DATE"] = Time.now().utc.iso
    _add_header_cards(header, **header_cards)

    return BinTableHDU(psf_, header=header, name=extname)


@u.quantity_input(
    true_energy_bins=u.TeV, fov_offset_bins=u.deg,
)
def create_energy_dispersion_hdu(
    energy_dispersion,
    true_energy_bins,
    migration_bins,
    fov_offset_bins,
    point_like=True,
    extname="EDISP",
    **header_cards,
):
    """
    Create a fits binary table HDU in GADF format for the energy dispersion.
    See the specification at
    https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/full_enclosure/edisp/index.html

    Parameters
    ----------
    energy_dispersion: numpy.ndarray
        Energy dispersion array, must have shape
        (n_energy_bins, n_migra_bins, n_source_offset_bins)
    true_energy_bins: astropy.units.Quantity[energy]
        Bin edges in true energy
    migration_bins: numpy.ndarray
        Bin edges for the relative energy migration (``reco_energy / true_energy``)
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.
    point_like: bool
        If the provided effective area was calculated after applying a direction cut,
        pass ``True``, else ``False`` for a full-enclosure effective area.
    extname: str
        Name for BinTableHDU
    **header_cards
        Additional metadata to add to the header, use this to set e.g. TELESCOP or
        INSTRUME.
    """

    edisp = QTable()
    edisp["ENERG_LO"], edisp["ENERG_HI"] = binning.split_bin_lo_hi(true_energy_bins[np.newaxis, :].to(u.TeV))
    edisp["MIGRA_LO"], edisp["MIGRA_HI"] = binning.split_bin_lo_hi(migration_bins[np.newaxis, :])
    edisp["THETA_LO"], edisp["THETA_HI"] = binning.split_bin_lo_hi(fov_offset_bins[np.newaxis, :].to(u.deg))
    # transpose as FITS uses opposite dimension order
    edisp["MATRIX"] = u.Quantity(energy_dispersion.T[np.newaxis, ...]).to(u.one)

    # required header keywords
    header = DEFAULT_HEADER.copy()
    header["HDUCLAS1"] = "RESPONSE"
    header["HDUCLAS2"] = "EDISP"
    header["HDUCLAS3"] = "POINT-LIKE" if point_like else "FULL-ENCLOSURE"
    header["HDUCLAS4"] = "EDISP_2D"
    header["DATE"] = Time.now().utc.iso
    _add_header_cards(header, **header_cards)

    return BinTableHDU(edisp, header=header, name=extname)


#: Unit to store background rate in GADF format
#:
#: see https://github.com/open-gamma-ray-astro/gamma-astro-data-formats/issues/153
#: for a discussion on why this is MeV not TeV as everywhere else
GADF_BACKGROUND_UNIT = u.Unit("MeV-1 s-1 sr-1")


@u.quantity_input(
    background=GADF_BACKGROUND_UNIT, reco_energy_bins=u.TeV, fov_offset_bins=u.deg,
)
def create_background_2d_hdu(
    background_2d,
    reco_energy_bins,
    fov_offset_bins,
    extname="BACKGROUND",
    **header_cards,
):
    """
    Create a fits binary table HDU in GADF format for the background 2d table.
    See the specification at
    https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/full_enclosure/bkg/index.html#bkg-2d

    Parameters
    ----------
    background_2d: astropy.units.Quantity[(MeV s sr)^-1]
        Background rate, must have shape
        (n_energy_bins, n_fov_offset_bins)
    reco_energy_bins: astropy.units.Quantity[energy]
        Bin edges in reconstructed energy
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
    extname: str
        Name for BinTableHDU
    **header_cards
        Additional metadata to add to the header, use this to set e.g. TELESCOP or
        INSTRUME.
    """

    bkg = QTable()
    bkg["ENERG_LO"], bkg["ENERG_HI"] = binning.split_bin_lo_hi(reco_energy_bins[np.newaxis, :].to(u.TeV))
    bkg["THETA_LO"], bkg["THETA_HI"] = binning.split_bin_lo_hi(fov_offset_bins[np.newaxis, :].to(u.deg))
    # transpose as FITS uses opposite dimension order
    bkg["BKG"] = background_2d.T[np.newaxis, ...].to(GADF_BACKGROUND_UNIT)

    # required header keywords
    header = DEFAULT_HEADER.copy()
    header["HDUCLAS1"] = "RESPONSE"
    header["HDUCLAS2"] = "BKG"
    header["HDUCLAS3"] = "FULL-ENCLOSURE"
    header["HDUCLAS4"] = "BKG_2D"
    header["DATE"] = Time.now().utc.iso
    _add_header_cards(header, **header_cards)

    return BinTableHDU(bkg, header=header, name=extname)


@u.quantity_input(
    rad_max=u.deg,
    reco_energy_bins=u.TeV,
    fov_offset_bins=u.deg,
)
def create_rad_max_hdu(
    rad_max,
    reco_energy_bins,
    fov_offset_bins,
    point_like=True,
    extname="RAD_MAX",
    **header_cards,
):
    """
    Create a fits binary table HDU in GADF format for the directional cut.
    See the specification at
    https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/point_like/index.html#rad-max

    Parameters
    ----------
    rad_max: astropy.units.Quantity[angle]
        Array of the directional (theta) cut.
        Must have shape (n_reco_energy_bins, n_fov_offset_bins)
    reco_energy_bins: astropy.units.Quantity[energy]
        Bin edges in reconstructed energy
    fov_offset_bins: astropy.units.Quantity[angle]
        Bin edges in the field of view offset.
        For Point-Like IRFs, only giving a single bin is appropriate.
    extname: str
        Name for BinTableHDU
    **header_cards
        Additional metadata to add to the header, use this to set e.g. TELESCOP or
        INSTRUME.
    """

    rad_max_table = QTable()
    rad_max_table["ENERG_LO"], rad_max_table["ENERG_HI"] = binning.split_bin_lo_hi(reco_energy_bins[np.newaxis, :].to(u.TeV))
    rad_max_table["THETA_LO"], rad_max_table["THETA_HI"] = binning.split_bin_lo_hi(fov_offset_bins[np.newaxis, :].to(u.deg))
    # transpose as FITS uses opposite dimension order
    rad_max_table["RAD_MAX"] = rad_max.T[np.newaxis, ...].to(u.deg)

    # required header keywords
    header = DEFAULT_HEADER.copy()
    header["HDUCLAS1"] = "RESPONSE"
    header["HDUCLAS2"] = "RAD_MAX"
    header["HDUCLAS3"] = "POINT-LIKE"
    header["HDUCLAS4"] = "RAD_MAX_2D"
    header["DATE"] = Time.now().utc.iso
    _add_header_cards(header, **header_cards)

    return BinTableHDU(rad_max_table, header=header, name=extname)



import astropy.units as u
from astropy.table import QTable
import numpy as np


def test_background():
    from pyirf.irf import background_2d
    from pyirf.utils import cone_solid_angle

    np.random.seed(0)

    N1 = 1000
    N2 = 100
    N = N1 + N2

    # toy event data set with just two energies
    # and a psf per energy bin, point-like
    events = QTable(
        {
            "reco_energy": np.append(np.full(N1, 1), np.full(N2, 2)) * u.TeV,
            "reco_source_fov_offset": np.zeros(N) * u.deg,
            "weight": np.ones(N),
        }
    )

    energy_bins = [0, 1.5, 3] * u.TeV
    fov_bins = [0, 1] * u.deg

    # We return a table with one row as needed for gadf
    bg = background_2d(events, energy_bins, fov_bins, t_obs=1 * u.s)

    # 2 energy bins, 1 fov bin, 200 source distance bins
    assert bg.shape == (2, 1)
    assert bg.unit == u.Unit("TeV-1 s-1 sr-1")

    # check that psf is normalized
    bin_solid_angle = np.diff(cone_solid_angle(fov_bins))
    e_width = np.diff(energy_bins)
    assert np.allclose(np.sum((bg.T * e_width).T * bin_solid_angle, axis=1), [1000, 100] / u.s)



from .eventdisplay import read_eventdisplay_fits
from .gadf import (
    create_aeff2d_hdu,
    create_energy_dispersion_hdu,
    create_psf_table_hdu,
    create_rad_max_hdu,
    create_background_2d_hdu,
)


__all__ = [
    "read_eventdisplay_fits",
    "create_psf_table_hdu",
    "create_aeff2d_hdu",
    "create_energy_dispersion_hdu",
    "create_psf_table_hdu",
    "create_rad_max_hdu",
    "create_background_2d_hdu",
]



import numpy as np
from astropy.table import QTable
from scipy.stats import norm
import astropy.units as u

from ..binning import calculate_bin_indices


ONE_SIGMA_QUANTILE = norm.cdf(1) - norm.cdf(-1)


def angular_resolution(
    events, energy_bins, energy_type="true",
):
    """
    Calculate the angular resolution.

    This implementation corresponds to the 68% containment of the angular
    distance distribution.

    Parameters
    ----------
    events : astropy.table.QTable
        Astropy Table object containing the reconstructed events information.
    energy_bins: numpy.ndarray(dtype=float, ndim=1)
        Bin edges in energy.
    energy_type: str
        Either "true" or "reco" energy.
        Default is "true".

    Returns
    -------
    result : astropy.table.QTable
        QTable containing the 68% containment of the angular
        distance distribution per each reconstructed energy bin.
    """
    # create a table to make use of groupby operations
    energy_key = f"{energy_type}_energy"
    table = QTable(events[[energy_key, "theta"]])

    bin_index, valid = calculate_bin_indices(table[energy_key], energy_bins)

    result = QTable()
    result[f"{energy_key}_low"] = energy_bins[:-1]
    result[f"{energy_key}_high"] = energy_bins[1:]
    result[f"{energy_key}_center"] = 0.5 * (energy_bins[:-1] + energy_bins[1:])
    result["n_events"] = 0

    key = "angular_resolution"
    result[key] = u.Quantity(np.nan, table["theta"].unit)

    # if we get an empty input (no selected events available)
    # we return the table filled with NaNs
    if len(events) == 0:
        return result

    # use groupby operations to calculate the percentile in each bin
    by_bin = table[valid].group_by(bin_index[valid])
    for bin_idx, group in zip(by_bin.groups.keys, by_bin.groups):
        result[key][bin_idx] = np.nanquantile(group["theta"], ONE_SIGMA_QUANTILE)
    return result



import numpy as np
import astropy.units as u
from astropy.coordinates.angle_utilities import angular_separation

from .exceptions import MissingColumns, WrongColumnUnit


__all__ = [
    "is_scalar",
    "calculate_theta",
    "calculate_source_fov_offset",
    "check_histograms",
    "cone_solid_angle",
]


def is_scalar(val):
    """Workaround that also supports astropy quantities

    Parameters
    ----------
    val : object
        Any object (value, list, etc...)

    Returns
    -------
    result: bool
        True is if input object is a scalar, False otherwise.
    """
    result = np.array(val, copy=False).shape == tuple()
    return result


@u.quantity_input(assumed_source_az=u.deg, assumed_source_alt=u.deg)
def calculate_theta(events, assumed_source_az, assumed_source_alt):
    """Calculate sky separation between assumed and reconstructed positions.

    Parameters
    ----------
    events : astropy.QTable
        Astropy Table object containing the reconstructed events information.
    assumed_source_az: astropy.units.Quantity
        Assumed Azimuth angle of the source.
    assumed_source_alt: astropy.units.Quantity
        Assumed Altitude angle of the source.

    Returns
    -------
    theta: astropy.units.Quantity
        Angular separation between the assumed and reconstructed positions
        in the sky.
    """
    theta = angular_separation(
        assumed_source_az,
        assumed_source_alt,
        events["reco_az"],
        events["reco_alt"],
    )

    return theta.to(u.deg)


def calculate_source_fov_offset(events, prefix="true"):
    """Calculate angular separation between true and pointing positions.

    Parameters
    ----------
    events : astropy.QTable
        Astropy Table object containing the reconstructed events information.

    prefix: str
        Column prefix for az / alt, can be used to calculate reco or true
        source fov offset.

    Returns
    -------
    theta: astropy.units.Quantity
        Angular separation between the true and pointing positions
        in the sky.
    """
    theta = angular_separation(
        events[f"{prefix}_az"],
        events[f"{prefix}_alt"],
        events["pointing_az"],
        events["pointing_alt"],
    )

    return theta.to(u.deg)


def check_histograms(hist1, hist2, key="reco_energy"):
    """
    Check if two histogram tables have the same binning

    Parameters
    ----------
    hist1: ``~astropy.table.Table``
        First histogram table, as created by
        ``~pyirf.binning.create_histogram_table``
    hist2: ``~astropy.table.Table``
        Second histogram table
    """

    # check binning information and add to output
    for k in ("low", "center", "high"):
        k = key + "_" + k
        if not np.all(hist1[k] == hist2[k]):
            raise ValueError(
                "Binning for signal_hist and background_hist must be equal"
            )


def cone_solid_angle(angle):
    """Calculate the solid angle of a view cone.

    Parameters
    ----------
    angle: astropy.units.Quantity or astropy.coordinates.Angle
        Opening angle of the view cone.

    Returns
    -------
    solid_angle: astropy.units.Quantity
        Solid angle of a view cone with opening angle ``angle``.

    """
    solid_angle = 2 * np.pi * (1 - np.cos(angle)) * u.sr
    return solid_angle


def check_table(table, required_columns=None, required_units=None):
    """Check a table for required columns and units.

    Parameters
    ----------
    table: astropy.table.QTable
        Table to check
    required_columns: iterable[str]
        Column names that are required to be present
    required_units: Mapping[str->astropy.units.Unit]
        Required units for columns as a Mapping from column names to units.
        Checks if the units are convertible, not if they are identical

    Raises
    ------
    MissingColumns: If any of the columns specified in ``required_columns`` or
        as keys in ``required_units are`` not present in the table.
    WrongColumnUnit: if any column has the wrong unit
    """
    if required_columns is not None:
        missing = set(required_columns) - set(table.colnames)
        if missing:
            raise MissingColumns(missing)

    if required_units is not None:
        for col, expected in required_units.items():
            if col not in table.colnames:
                raise MissingColumns(col)

            unit = table[col].unit
            if not expected.is_equivalent(unit):
                raise WrongColumnUnit(col, unit, expected)



import numpy as np
import astropy.units as u
from ..binning import create_histogram_table


__all__ = [
    "effective_area",
    "effective_area_per_energy",
    "effective_area_per_energy_and_fov",
]


@u.quantity_input(area=u.m ** 2)
def effective_area(n_selected, n_simulated, area):
    """
    Calculate effective area for histograms of selected and total simulated events

    Parameters
    ----------
    n_selected: int or numpy.ndarray[int]
        The number of surviving (e.g. triggered, analysed, after cuts)
    n_simulated: int or numpy.ndarray[int]
        The total number of events simulated
    area: astropy.units.Quantity[area]
        Area in which particle's core position was simulated
    """
    return (n_selected / n_simulated) * area


def effective_area_per_energy(selected_events, simulation_info, true_energy_bins):
    """
    Calculate effective area in bins of true energy.

    Parameters
    ----------
    selected_events: astropy.table.QTable
        DL2 events table, required columns for this function: `true_energy`.
    simulation_info: pyirf.simulations.SimulatedEventsInfo
        The overall statistics of the simulated events
    true_energy_bins: astropy.units.Quantity[energy]
        The bin edges in which to calculate effective area.
    """
    area = np.pi * simulation_info.max_impact ** 2

    hist_selected = create_histogram_table(
        selected_events, true_energy_bins, "true_energy"
    )
    hist_simulated = simulation_info.calculate_n_showers_per_energy(true_energy_bins)

    return effective_area(hist_selected["n"], hist_simulated, area)


def effective_area_per_energy_and_fov(
    selected_events, simulation_info, true_energy_bins, fov_offset_bins
):
    """
    Calculate effective area in bins of true energy and field of view offset.

    Parameters
    ----------
    selected_events: astropy.table.QTable
        DL2 events table, required columns for this function:
        - `true_energy`
        - `true_source_fov_offset`
    simulation_info: pyirf.simulations.SimulatedEventsInfo
        The overall statistics of the simulated events
    true_energy_bins: astropy.units.Quantity[energy]
        The true energy bin edges in which to calculate effective area.
    fov_offset_bins: astropy.units.Quantity[angle]
        The field of view radial bin edges in which to calculate effective area.
    """
    area = np.pi * simulation_info.max_impact ** 2

    hist_simulated = simulation_info.calculate_n_showers_per_energy_and_fov(
        true_energy_bins, fov_offset_bins
    )

    hist_selected, _, _ = np.histogram2d(
        selected_events["true_energy"].to_value(u.TeV),
        selected_events["true_source_fov_offset"].to_value(u.deg),
        bins=[
            true_energy_bins.to_value(u.TeV),
            fov_offset_bins.to_value(u.deg),
        ],
    )

    return effective_area(hist_selected, hist_simulated, area)



from astropy.table import QTable
import astropy.units as u
import numpy as np
import pytest


def test_empty_angular_resolution():
    from pyirf.benchmarks import angular_resolution

    events = QTable({
        'true_energy': [] * u.TeV,
        'theta': [] * u.deg,
    })

    table = angular_resolution(
        events,
        [1, 10, 100] * u.TeV
    )

    assert np.all(np.isnan(table["angular_resolution"]))

@pytest.mark.parametrize("unit", (u.deg, u.rad))
def test_angular_resolution(unit):
    from pyirf.benchmarks import angular_resolution

    np.random.seed(1337)

    TRUE_RES_1 = 0.2
    TRUE_RES_2 = 0.05
    true_resolution = np.append(np.full(1000, TRUE_RES_1), np.full(1000, TRUE_RES_2))

    events = QTable({
        'true_energy': np.concatenate([
            [0.5], # below bin 1 to test with underflow
            np.full(999, 5.0),
            np.full(999, 50.0),
            [500], # above bin 2 to test with overflow
        ]) * u.TeV,
        'theta': np.abs(np.random.normal(0, true_resolution)) * u.deg
    })

    events['theta'] = events['theta'].to(unit)

    # add nans to test if nans are ignored
    events["true_energy"].value[500] = np.nan
    events["true_energy"].value[1500] = np.nan

    ang_res = angular_resolution(
        events,
        [1, 10, 100] * u.TeV,
    )['angular_resolution']

    assert len(ang_res) == 2
    assert u.isclose(ang_res[0], TRUE_RES_1 * u.deg, rtol=0.05)
    assert u.isclose(ang_res[1], TRUE_RES_2 * u.deg, rtol=0.05)



import astropy.units as u
import numpy as np

__all__ = [
    'SimulatedEventsInfo',
]


class SimulatedEventsInfo:
    """
    Information about all simulated events, for calculating event weights.

    Attributes
    ----------
    n_showers: int
        Total number of simulated showers. If reuse was used, this
        should already include the reuse.
    energy_min: u.Quantity[energy]
        Lower limit of the simulated energy range
    energy_max: u.Quantity[energy]
        Upper limit of the simulated energy range
    max_impact: u.Quantity[length]
        Maximum simulated impact parameter
    spectral_index: float
        Spectral Index of the simulated power law with sign included.
    viewcone: u.Quantity[angle]
        Opening angle of the viewcone
    """

    __slots__ = (
        "n_showers",
        "energy_min",
        "energy_max",
        "max_impact",
        "spectral_index",
        "viewcone",
    )

    @u.quantity_input(
        energy_min=u.TeV, energy_max=u.TeV, max_impact=u.m, viewcone=u.deg
    )
    def __init__(
        self, n_showers, energy_min, energy_max, max_impact, spectral_index, viewcone
    ):
        #: Total number of simulated showers, if reuse was used, this must
        #: already include reuse
        self.n_showers = n_showers
        #: Lower limit of the simulated energy range
        self.energy_min = energy_min
        #: Upper limit of the simulated energy range
        self.energy_max = energy_max
        #: Maximum simualted impact radius
        self.max_impact = max_impact
        #: Spectral index of the simulated power law with sign included
        self.spectral_index = spectral_index
        #: Opening angle of the viewcone
        self.viewcone = viewcone

        if spectral_index > -1:
            raise ValueError("spectral index must be <= -1")

    @u.quantity_input(energy_bins=u.TeV)
    def calculate_n_showers_per_energy(self, energy_bins):
        """
        Calculate number of showers that were simulated in the given energy intervals

        This assumes the events were generated and from a powerlaw
        like CORSIKA simulates events.

        Parameters
        ----------
        energy_bins: astropy.units.Quantity[energy]
            The interval edges for which to calculate the number of simulated showers

        Returns
        -------
        n_showers: numpy.ndarray
            The expected number of events inside each of the ``energy_bins``.
            This is a floating point number.
            The actual numbers will follow a poissionian distribution around this
            expected value.
        """
        bins = energy_bins
        e_low = bins[:-1]
        e_high = bins[1:]
        e_min = self.energy_min
        e_max = self.energy_max

        integral = _powerlaw_pdf_integral(
            self.spectral_index, e_low, e_high, e_min, e_max
        )

        integral[e_high <= e_min] = 0
        integral[e_low >= e_max] = 0

        mask = (e_high > e_max) & (e_low < e_max)
        integral[mask] = _powerlaw_pdf_integral(
            self.spectral_index, e_low[mask], e_max, e_min, e_max
        )

        mask = (e_high > e_min) & (e_low < e_min)
        integral[mask] = _powerlaw_pdf_integral(
            self.spectral_index, e_min, e_high[mask], e_min, e_max
        )

        return self.n_showers * integral

    def calculate_n_showers_per_fov(self, fov_bins):
        """
        Calculate number of showers that were simulated in the given fov bins.

        This assumes the events were generated uniformly distributed per solid angle,
        like CORSIKA simulates events with the VIEWCONE option.

        Parameters
        ----------
        fov_bins: astropy.units.Quantity[angle]
            The FOV bin edges for which to calculate the number of simulated showers

        Returns
        -------
        n_showers: numpy.ndarray(ndim=2)
            The expected number of events inside each of the ``fov_bins``.
            This is a floating point number.
            The actual numbers will follow a poissionian distribution around this
            expected value.
        """
        fov_bins = fov_bins
        fov_low = fov_bins[:-1]
        fov_high = fov_bins[1:]

        fov_integral = _viewcone_pdf_integral(self.viewcone, fov_low, fov_high)
        viewcone = self.viewcone
        # check if any of the bins are outside the max viewcone
        fov_integral = np.where(fov_high <= viewcone, fov_integral, 0)

        # identify the bin with the maximum viewcone inside
        mask = (viewcone > fov_low) & (viewcone < fov_high)
        fov_integral[mask] = _viewcone_pdf_integral(viewcone, fov_low[mask], viewcone)

        return self.n_showers * fov_integral

    @u.quantity_input(energy_bins=u.TeV, fov_bins=u.deg)
    def calculate_n_showers_per_energy_and_fov(self, energy_bins, fov_bins):
        """
        Calculate number of showers that were simulated in the given
        energy and fov bins.

        This assumes the events were generated uniformly distributed per solid angle,
        and from a powerlaw in energy like CORSIKA simulates events.

        Parameters
        ----------
        energy_bins: astropy.units.Quantity[energy]
            The energy bin edges for which to calculate the number of simulated showers
        fov_bins: astropy.units.Quantity[angle]
            The FOV bin edges for which to calculate the number of simulated showers

        Returns
        -------
        n_showers: numpy.ndarray(ndim=2)
            The expected number of events inside each of the
            ``energy_bins`` and ``fov_bins``.
            Dimension (n_energy_bins, n_fov_bins)
            This is a floating point number.
            The actual numbers will follow a poissionian distribution around this
            expected value.
        """
        # energy distribution and fov distribution are independent in CORSIKA,
        # so just multiply both distributions.
        e_integral = self.calculate_n_showers_per_energy(energy_bins)
        fov_integral = self.calculate_n_showers_per_fov(fov_bins)
        return e_integral[:, np.newaxis] * fov_integral / self.n_showers

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"n_showers={self.n_showers}, "
            f"energy_min={self.energy_min:.3f}, "
            f"energy_max={self.energy_max:.2f}, "
            f"spectral_index={self.spectral_index:.1f}, "
            f"max_impact={self.max_impact:.2f}, "
            f"viewcone={self.viewcone}"
            ")"
        )


def _powerlaw_pdf_integral(index, e_low, e_high, e_min, e_max):
    # strip units, make sure all in the same unit
    e_low = e_low.to_value(u.TeV)
    e_high = e_high.to_value(u.TeV)
    e_min = e_min.to_value(u.TeV)
    e_max = e_max.to_value(u.TeV)

    int_index = index + 1
    normalization = 1 / (e_max ** int_index - e_min ** int_index)
    e_term = e_high ** int_index - e_low ** int_index
    return e_term * normalization


def _viewcone_pdf_integral(viewcone, fov_low, fov_high):
    if viewcone.value == 0:
        raise ValueError("Only supported for diffuse simulations")
    else:
        norm = 1 / (1 - np.cos(viewcone))

    integral = np.cos(fov_low) - np.cos(fov_high)

    return norm * integral



import numpy as np
from astropy.table import Table, QTable
from scipy.ndimage import gaussian_filter1d
import astropy.units as u

from .binning import calculate_bin_indices, bin_center

__all__ = [
    'calculate_percentile_cut',
    'evaluate_binned_cut',
    'compare_irf_cuts',
]


def calculate_percentile_cut(
    values, bin_values, bins, fill_value, percentile=68, min_value=None, max_value=None,
    smoothing=None, min_events=10,
):
    """
    Calculate cuts as the percentile of a given quantity in bins of another
    quantity.

    Parameters
    ----------
    values: ``~numpy.ndarray`` or ``~astropy.units.Quantity``
        The values for which the cut should be calculated
    bin_values: ``~numpy.ndarray`` or ``~astropy.units.Quantity``
        The values used to sort the ``values`` into bins
    edges: ``~numpy.ndarray`` or ``~astropy.units.Quantity``
        Bin edges
    fill_value: float or quantity
        Value for bins with less than ``min_events``,
        must have same unit as values
    percentile: float
        The percentile to calculate in each bin as a percentage,
        i.e. 0 <= percentile <= 100.
    min_value: float or quantity or None
        If given, cuts smaller than this value are replaced with ``min_value``
    max_value: float or quantity or None
        If given, cuts larger than this value are replaced with ``max_value``
    smoothing: float or None
        If given, apply a gaussian filter of width ``sigma`` in terms
        of bins.
    min_events: int
        Bins with less events than this number are replaced with ``fill_value``
    """
    # create a table to make use of groupby operations
    # we use a normal table here to avoid astropy/astropy#13840
    table = Table({"values": values}, copy=False)
    unit = table["values"].unit

    # make sure units match
    if unit is not None:
        fill_value = u.Quantity(fill_value).to(unit)

        if min_value is not None:
            min_value = u.Quantity(min_value).to_value(unit)

        if max_value is not None:
            max_value = u.Quantity(max_value).to_value(unit)

    bin_index, valid = calculate_bin_indices(bin_values, bins)
    by_bin = table[valid].group_by(bin_index[valid])

    cut_table = QTable()
    cut_table["low"] = bins[:-1]
    cut_table["high"] = bins[1:]
    cut_table["center"] = bin_center(bins)
    cut_table["n_events"] = 0
    cut_table["cut"] = np.asanyarray(fill_value, values.dtype)

    for bin_idx, group in zip(by_bin.groups.keys, by_bin.groups):
        # replace bins with too few events with fill_value
        n_events = len(group)
        cut_table["n_events"][bin_idx] = n_events

        if n_events < min_events:
            cut_table["cut"][bin_idx] = fill_value
        else:
            value = np.nanpercentile(group["values"], percentile)
            if min_value is not None or max_value is not None:
                value = np.clip(value, min_value, max_value)

            cut_table["cut"].value[bin_idx] = value

    if smoothing is not None:
        cut_table['cut'].value[:] = gaussian_filter1d(
            cut_table["cut"].value,
            smoothing,
            mode='nearest',
        )

    return cut_table


def evaluate_binned_cut(values, bin_values, cut_table, op):
    """
    Evaluate a binned cut as defined in cut_table on given events.

    Events with bin_values outside the bin edges defined in cut table
    will be set to False.

    Parameters
    ----------
    values: ``~numpy.ndarray`` or ``~astropy.units.Quantity``
        The values on which the cut should be evaluated
    bin_values: ``~numpy.ndarray`` or ``~astropy.units.Quantity``
        The values used to sort the ``values`` into bins
    cut_table: ``~astropy.table.Table``
        A table describing the binned cuts, e.g. as created by
        ``~pyirf.cuts.calculate_percentile_cut``.
        Required columns:
        - `low`: lower edges of the bins
        - `high`: upper edges of the bins,
        - `cut`: cut value
    op: callable(a, b) -> bool
        A function taking two arguments, comparing element-wise and
        returning an array of booleans.
        Must support vectorized application.
    """
    if not isinstance(cut_table, QTable):
        raise ValueError('cut_table needs to be an astropy.table.QTable')

    bins = np.append(cut_table["low"], cut_table["high"][-1])
    bin_index, valid = calculate_bin_indices(bin_values, bins)

    result = np.zeros(len(values), dtype=bool)
    result[valid] = op(values[valid], cut_table["cut"][bin_index[valid]])
    return result


def compare_irf_cuts(cuts):
    """
    checks if the same cuts have been applied in all of them

    Parameters
    ----------
    cuts: list of QTables
        list of cuts each entry in the list correspond to one set of IRFs
    Returns
    -------
    match: Boolean
        if the cuts are the same in all the files
    """
    for i in range(len(cuts) - 1):
        if (cuts[i] != cuts[i + 1]).any():
            return False
    return True



# coding: utf-8
import tarfile
from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

import requests


def download_irfs():
    r = requests.get(
        "https://zenodo.org/record/5499840/files/cta-prod5-zenodo-fitsonly-v0.1.zip?download=1"
    )
    r.raise_for_status()

    obstime = 50 * 3600

    tarball = (
        "fits/CTA-Performance-prod5-v0.1-North-LSTSubArray-{zenith:d}deg.FITS.tar.gz"
    )
    irf_file = "Prod5-North-{zenith:d}deg-AverageAz-4LSTs.{obstime}s-v0.1.fits.gz"

    output_dir = Path(__file__).parent / "irfs"
    output_dir.mkdir(exist_ok=True)

    for zenith in (20, 40, 60):
        with ZipFile(BytesIO(r.content)) as zipfile:
            with tarfile.open(
                fileobj=zipfile.open(tarball.format(zenith=zenith), mode="r")
            ) as tar:
                tar.extract(
                    irf_file.format(zenith=zenith, obstime=obstime), path=output_dir
                )


if __name__ == "__main__":
    download_irfs()



import astropy.units as u
from astropy.table import QTable
import numpy as np
import pyirf.binning as binning


def test_add_overflow_bins():
    from pyirf.binning import add_overflow_bins

    bins = np.array([1, 2])
    bins_uo = add_overflow_bins(bins)

    assert len(bins_uo) == 4
    assert bins_uo[0] == 0
    assert np.isinf(bins_uo[-1])
    assert np.all(bins_uo[1:-1] == bins)

    bins = np.array([1, 2])
    bins_uo = add_overflow_bins(bins, positive=False)

    assert bins_uo[0] < 0
    assert np.isinf(bins_uo[0])

    # make sure we don't any bins if over / under is already present
    bins = np.array([0, 1, 2, np.inf])
    assert len(add_overflow_bins(bins)) == len(bins)


def test_add_overflow_bins_units():
    from pyirf.binning import add_overflow_bins

    bins = np.array([1, 2]) * u.TeV
    bins_uo = add_overflow_bins(bins)

    assert bins_uo.unit == bins.unit
    assert len(bins_uo) == 4
    assert bins_uo[0] == 0 * u.TeV
    assert np.isinf(bins_uo[-1])
    assert np.all(bins_uo[1:-1] == bins)


def test_bins_per_decade():
    from pyirf.binning import create_bins_per_decade

    bins = create_bins_per_decade(100 * u.GeV, 100 * u.TeV)

    assert bins.unit == u.GeV
    assert len(bins) == 16  # end inclusive if exactly fits per_decade

    assert bins[0] == 100 * u.GeV
    assert np.allclose(np.diff(np.log10(bins.to_value(u.GeV))), 0.2)

    bins = create_bins_per_decade(100 * u.GeV, 100 * u.TeV, 10)
    assert bins.unit == u.GeV
    assert len(bins) == 31  # end inclusive since it fits exactly

    assert bins[0] == 100 * u.GeV
    assert np.allclose(np.diff(np.log10(bins.to_value(u.GeV))), 0.1)

    bins = create_bins_per_decade(100 * u.GeV, 105 * u.TeV, 5)
    assert bins.unit == u.GeV
    assert len(bins) == 16  # end non-inclusive

    assert u.isclose(bins[-1], 100 * u.TeV) # last value at correct difference
    assert np.allclose(np.diff(np.log10(bins.to_value(u.GeV))), 0.2)


def test_create_histogram_table():
    '''Test create histogram table'''
    from pyirf.binning import create_histogram_table

    events = QTable({
        'reco_energy': [1, 1, 10, 100, 100, 100] * u.TeV,
    })
    bins = [0.5, 5, 50, 500] * u.TeV

    # test without weights
    hist = create_histogram_table(events, bins, key='reco_energy')
    assert np.all(hist['n'] == [2, 1, 3])
    assert np.all(hist['n_weighted'] == [2, 1, 3])

    # test with weights
    events['weight'] = [0.5, 2, 2.5, 0.5, 0.2, 0.3]
    hist = create_histogram_table(events, bins, key='reco_energy')
    assert np.all(hist['n'] == [2, 1, 3])
    assert np.all(hist['n_weighted'] == [2.5, 2.5, 1.0])

    # test with particle_types
    types = np.array(['gamma', 'electron', 'gamma', 'electron', 'gamma', 'proton'])
    events['particle_type'] = types
    hist = create_histogram_table(events, bins, key='reco_energy')

    assert np.all(hist['n'] == [2, 1, 3])
    assert np.all(hist['n_weighted'] == [2.5, 2.5, 1.0])

    assert np.all(hist['n_gamma'] == [1, 1, 1])
    assert np.all(hist['n_electron'] == [1, 0, 1])
    assert np.all(hist['n_proton'] == [0, 0, 1])

    assert np.allclose(hist['n_gamma_weighted'], [0.5, 2.5, 0.2])
    assert np.allclose(hist['n_electron_weighted'], [2, 0, 0.5])
    assert np.allclose(hist['n_proton_weighted'], [0, 0, 0.3])

    # test with empty table
    empty = events[:0]
    hist = create_histogram_table(empty, bins, key='reco_energy')
    zeros = np.zeros(3)
    assert np.all(hist['n'] == zeros)
    assert np.all(hist['n_weighted'] == zeros)


def test_calculate_bin_indices():
    from pyirf.binning import calculate_bin_indices, OVERFLOW_INDEX, UNDERFLOW_INDEX

    bins = np.array([0, 1, 2])
    values = [0.5, 0.5, 1, 1.1, 1.9, 2, -1, 2.5]

    true_idx = np.array([0, 0, 1, 1, 1, 2, -1, 2])
    true_valid = np.array([True, True, True, True, True, False, False, False])
    true_idx = np.array([0, 0, 1, 1, 1, OVERFLOW_INDEX, UNDERFLOW_INDEX, OVERFLOW_INDEX])

    idx, valid = calculate_bin_indices(values, bins)
    assert np.all(idx == true_idx)
    assert np.all(valid == true_valid)

    # test with units
    bins *= u.TeV
    values *= 1000 * u.GeV
    idx, valid = calculate_bin_indices(values, bins)
    assert np.all(idx == true_idx)
    assert np.all(valid == true_valid)


def test_resample_bins_1d():
    from pyirf.binning import resample_histogram1d

    n = 10
    data = np.ones(n)
    edges = np.arange(n + 1)

    # no resampling
    new_data = resample_histogram1d(data, edges, edges)
    np.testing.assert_array_almost_equal(new_data, data)

    # resampling, less bins
    new_edges = np.arange(n + 1, step=2)
    true_new_data = np.ones(n // 2) * 2
    new_data = resample_histogram1d(data, edges, new_edges)
    np.testing.assert_array_almost_equal(new_data, true_new_data)

    # resampling, smaller range (lose normalization)
    new_edges = np.arange(n // 2 + 1)
    true_new_data = np.ones(n // 2)
    new_data = resample_histogram1d(data, edges, new_edges)
    np.testing.assert_array_almost_equal(new_data, true_new_data)

    # resampling, larger range (fill with zeros)
    new_edges = np.arange(-n, 2 * n + 1)
    true_new_data = np.concatenate([np.zeros(n), np.ones(n), np.zeros(n)])
    new_data = resample_histogram1d(data, edges, new_edges)
    np.testing.assert_array_almost_equal(new_data, true_new_data)


def test_resample_bins_nd():
    from pyirf.binning import resample_histogram1d
    n = 10

    # shape 2x3x10
    data = np.array([
        [
            np.ones(n),
            np.ones(n) * 2,
            np.ones(n) * 3,
        ],
        [
            np.ones(n) * 4,
            np.ones(n) * 5,
            np.ones(n) * 6,
        ],
    ])

    edges = np.arange(n + 1)
    new_data = resample_histogram1d(data, edges, edges, axis=2)
    np.testing.assert_array_almost_equal(new_data, data)

    edges = np.arange(3 + 1)
    new_data = resample_histogram1d(data, edges, edges, axis=1)
    np.testing.assert_array_almost_equal(new_data, data)


def test_resample_bins_units():
    from pyirf.binning import resample_histogram1d
    import pytest

    n = 10
    data = np.ones(n) * u.m

    old_edges = np.arange(n + 1) * u.TeV
    new_edges = old_edges.copy().to(u.GeV)

    new_data = resample_histogram1d(data, old_edges, new_edges)

    np.testing.assert_array_almost_equal(
        new_data.to_value(data.unit), data.to_value(data.unit),
    )

    with pytest.raises(ValueError):
        new_data = resample_histogram1d(data, old_edges, new_edges.value)

    with pytest.raises(u.core.UnitConversionError):
        new_data = resample_histogram1d(data, old_edges, u.Quantity(new_edges.value, unit=u.m))


def test_join_bin_lo_hi():
    """Test join_bin_hi_lo function."""
    bins = np.array([np.logspace(-1,3, 20)*u.TeV])
    bins_lo = bins[:,:-1]
    bins_hi = bins[:,1:]

    bins_joint = binning.join_bin_lo_hi(bins_lo,bins_hi)
    assert np.allclose(bins_joint, bins, rtol=1.e-5)


def test_split_bin_lo_hi():
    """Test split_bin_lo_hi function."""
    bins = np.array([np.logspace(-1,3, 20)*u.TeV])
    bins_lo_true = bins[:,:-1]
    bins_hi_true = bins[:,1:]

    bin_lo, bin_hi = binning.split_bin_lo_hi(bins)
    assert np.allclose(bin_lo, bins_lo_true, rtol=1.e-5)
    assert np.allclose(bin_hi, bins_hi_true, rtol=1.e-5)



import astropy.units as u
import numpy as np
from scipy.stats import expon

import pyirf.interpolation as interp
from pyirf.utils import cone_solid_angle


def test_interpolate_energy_dispersion(prod5_irfs):
    from pyirf.interpolation import interpolate_energy_dispersion

    zen_pnt = np.array([key.value for key in prod5_irfs.keys()])
    edisps = np.array([irf["edisp"].data for irf in prod5_irfs.values()])
    bin_edges = list(prod5_irfs.values())[0]["edisp"].axes["migra"].edges

    interp = interpolate_energy_dispersion(
        migra_bins=bin_edges,
        edisps=edisps[[0, 2]],
        grid_points=zen_pnt[[0, 2]],
        target_point=zen_pnt[[1]],
        quantile_resolution=1e-3,
    )

    assert np.max(interp) <= 1
    assert np.min(interp) >= 0
    assert np.all(np.isfinite(interp))
    assert np.all(
        np.logical_or(
            np.isclose(np.sum(interp, axis=-2), 1),
            np.isclose(np.sum(interp, axis=-2), 0),
        )
    )
    assert interp.shape == edisps[[1]].shape


def test_interpolate_psf_table():
    from pyirf.interpolation import interpolate_psf_table

    # dummy psf_table with 30 bins of true energ and 6 bins of fov-offset, rad-axis
    # to be inflated later
    dummy_psf_template = np.ones((30, 6, 1))

    zen_pnt = np.array([20, 40, 60])
    bin_edges = np.linspace(0, 1, 31) * u.deg
    omegas = np.diff(cone_solid_angle(bin_edges))

    def hist(pnt):
        """Create dummy psf for given pointing"""
        histogram = np.diff(expon(scale=pnt / 400).cdf(bin_edges))
        normed_hist = histogram / np.sum(histogram)

        return normed_hist / omegas

    dummy_psfs = np.array(
        [np.apply_along_axis(hist, -1, dummy_psf_template * pnt) for pnt in zen_pnt]
    )

    interp = interpolate_psf_table(
        source_offset_bins=bin_edges,
        psfs=dummy_psfs[[0, 2]],
        grid_points=zen_pnt[[0, 2]],
        target_point=zen_pnt[[1]],
        quantile_resolution=1e-3,
    )

    interp *= omegas[np.newaxis, np.newaxis, np.newaxis, ...]

    assert np.max(interp) <= 1
    assert np.min(interp) >= 0
    assert np.all(np.isfinite(interp))
    assert np.all(
        np.logical_or(
            np.isclose(np.sum(interp, axis=-1), 1),
            np.isclose(np.sum(interp, axis=-1), 0),
        )
    )
    assert interp.shape == dummy_psfs[[1]].shape


def test_interpolate_effective_area_per_energy_and_fov():
    """Test of interpolating of effective area using dummy model files."""
    n_en = 20
    n_th = 1
    en = np.logspace(-2, 2, n_en)
    # applying a simple sigmoid function
    aeff0 = 1.0e4 / (1 + 1 / en**2) * u.m**2

    # assume that for parameters 'x' and 'y' the Aeff scales x*y*Aeff0
    x = [0.9, 1.1]
    y = [8.0, 11.5]
    n_grid = len(x) * len(y)
    aeff = np.empty((n_grid, n_th, n_en))
    pars = np.empty((n_grid, 2))
    i_grid = 0
    for xx in x:
        for yy in y:
            aeff[i_grid, 0, :] = aeff0 * xx * yy / 10
            pars[i_grid, :] = np.array([xx, yy])
            i_grid += 1
    aeff *= u.m**2
    pars0 = np.array([1, 10])
    min_aeff = 1 * u.m**2
    aeff_interp = interp.interpolate_effective_area_per_energy_and_fov(
        aeff, pars, pars0, min_effective_area=min_aeff, method="linear"
    )

    # allowing for 3% accuracy except of close to the minimum value of Aeff
    assert np.allclose(aeff_interp[:, 0], aeff0, rtol=0.03, atol=min_aeff)


def test_interpolate_effective_area_per_energy_and_fov_prod5(prod5_irfs):
    """Test of interpolation of effective are on prod5 irfs"""
    from pyirf.interpolation import interpolate_effective_area_per_energy_and_fov

    zen_pnt = np.array([key.value for key in prod5_irfs.keys()])
    aeffs = np.array([irf["aeff"].data for irf in prod5_irfs.values()])

    interp = interpolate_effective_area_per_energy_and_fov(
        effective_area=aeffs[[0, 2]] * u.m**2,
        grid_points=zen_pnt[[0, 2]],
        target_point=zen_pnt[[1]],
        min_effective_area=1 * u.m**2,
    ).value

    assert np.all(np.isfinite(interp))
    assert interp.shape == aeffs[[1]].shape
    assert np.all(interp >= 0)

    assert np.all(
        np.logical_or(
            np.logical_or(
                np.logical_and(aeffs[[0]] <= interp, interp <= aeffs[[2]]),
                np.logical_and(aeffs[[2]] <= interp, interp <= aeffs[[0]]),
            ),
            interp == 0,
        )
    )


def test_interpolate_rad_max():
    from pyirf.interpolation import interpolate_rad_max

    # linear test case
    rad_max_1 = np.array([[0, 0], [1, 0], [2, 1], [3, 2]])
    rad_max_2 = 2 * rad_max_1
    rad_max = np.array([rad_max_1, rad_max_2])

    grid_points = np.array([[0], [1]])
    target_point = np.array([0.5])

    interp = interpolate_rad_max(
        rad_max=rad_max,
        grid_points=grid_points,
        target_point=target_point,
        method="linear",
    )

    assert interp.shape == (1, *rad_max_1.shape)
    assert np.allclose(interp, 1.5 * rad_max_1)



"""Base classes for interpolators"""
from abc import ABCMeta, abstractmethod

import numpy as np
from scipy.spatial import Delaunay

from pyirf.binning import bin_center

__all__ = ["BaseInterpolator", "ParametrizedInterpolator", "BinnedInterpolator"]


class BaseInterpolator(metaclass=ABCMeta):
    """
    Base class for all interpolators, only knowing grid-points,
    providing a common __call__-interface and doing sanity checks.
    """

    def __init__(self, grid_points):
        """BaseInterpolator constructor, doing sanity checks and ensuring
        correct shapes of inputs

        Parameters
        ----------
        grid_points: np.ndarray, shape=(n_points, n_dims):
            Grid points at which interpolation templates exist

        Raises
        ------
        TypeError:
            When grid_points is not a np.ndarray
        TypeError:
            When grid_point has dtype object
        ValueError:
            When there are too few points in grid_points to span a volume
            in the grid dimension.
        """
        if not isinstance(grid_points, np.ndarray):
            raise TypeError("Input grid_points is not a numpy array.")
        if grid_points.dtype == "O":
            raise TypeError("Input grid_points array cannot be of dtype object.")
        if not np.can_cast(grid_points.dtype, np.float128):
            raise TypeError("Input grid_points dtype incompatible with float.")

        self.grid_points = grid_points
        if self.grid_points.ndim == 1:
            self.grid_points = self.grid_points.reshape(*self.grid_points.shape, 1)
        self.n_points = self.grid_points.shape[0]
        self.grid_dim = self.grid_points.shape[1]

        # Check, if number of grid point theoretically suffices to span a volume
        # in the dimension indicated by grid
        if self.n_points < self.grid_dim + 1:
            raise ValueError(
                f"To few points for grid dimension, grid-dim is {self.grid_dim},"
                f" while there are only {self.n_points}. At least {self.grid_dim+1}"
                f" points needed to span a volume in {self.grid_dim} dimensions."
            )

        # Build triangulation to check if target is inside of the grid for
        # more then 1 dimension
        if self.grid_dim > 1:
            self.triangulation = Delaunay(self.grid_points)

    @abstractmethod
    def interpolate(self, target_point, **kwargs):
        """Overridable function for the actual interpolation code"""

    def _target_in_grid(self, target_point):
        """Check wether target_point lies within grids convex hull"""
        if self.grid_dim == 1:
            return (target_point >= self.grid_points.min()) and (
                target_point <= self.grid_points.max()
            )
        else:
            # Delaunay.find_simplex() returns -1 for points outside the grids convex hull
            simplex_ind = self.triangulation.find_simplex(target_point)
            return simplex_ind >= 0

    def __call__(self, target_point, extrapolator=None, **kwargs):
        """Providing a common __call__ interface sanity checking the target point

        Parameters
        ----------
        target_point: np.ndarray
            Target for inter-/extrapolation
        extrapolator: callable, optional
            Fall-Back extrapolator called when target_point is outside grid.
            Defaults to None, meaning no extrapolation is done.

        Raises
        ------
        TypeError:
            When target_point is not an np.ndarray
        ValueError:
            When more then one target_point is given
        ValueError:
            When target_point and grid_points have miss-matching dimensions
        ValueError:
            When target_point is outside of the grids convex hull but extrapolator is None

        Returns
        -------
        Interpolated or, if necessary extrapolated, result.
        """
        if not isinstance(target_point, np.ndarray):
            raise TypeError("Target point is not a numpy array.")

        if target_point.ndim == 1:
            target_point = target_point.reshape(1, *target_point.shape)
        elif target_point.shape[0] != 1:
            raise ValueError("Only one target_point per call supported.")

        if target_point.shape[1] != self.grid_dim:
            raise ValueError(
                "Missmatch between target-point and grid dimension."
                f" Grid has dimension {self.grid_dim}, target has dimension"
                f" {target_point.shape[1]}."
            )

        if self._target_in_grid(target_point):
            return self.interpolate(target_point, **kwargs)
        elif extrapolator is not None:
            print(f"Trying to extrapolate for point {target_point}.")
            return extrapolator(target_point, **kwargs)
        else:
            raise ValueError(
                "Target point outside grids convex hull and no extrapolator given."
            )


class ParametrizedInterpolator(BaseInterpolator):
    """
    Base class for all interpolators used with IRF components that can be
    independently interpolated, e.g. parametrized ones like 3Gauss
    but also AEff, extending BaseInterpolators sanity checks.
    Derived from pyirf.interpolation.BaseInterpolator
    """

    def __init__(self, grid_points, params):
        """ParametrizedInterpolator constructor

        Parameters
        ----------
        grid_points, np.ndarray
            Grid points at which interpolation templates exist
        params: np.ndarray
            Corresponding parameter values at each point in grid_points.
            First dimesion has to correspond to number of grid_points

        Raises
        ------
        TypeError:
            When params is not a np.ndarray
        ValueError:
            When number of points grid_points and params is not matching

        Note
        ----
            Also calls pyirf.interpolation.BaseInterpolators.__call__
        """
        super().__init__(grid_points)

        if not isinstance(params, np.ndarray):
            raise TypeError("Input params is not a numpy array.")
        elif self.n_points != params.shape[0]:
            raise ValueError(
                "Shape missmatch, number of grid_points and rows in params not matching."
            )
        else:
            self.params = params

        if self.params.ndim == 1:
            self.params = self.params[..., np.newaxis]


class BinnedInterpolator(BaseInterpolator):
    """
    Base class for all interpolators used with binned IRF components like EDisp,
    extending BaseInterpolators sanity checks.
    Derived from pyirf.interpolation.BaseInterpolator
    """

    def __init__(self, grid_points, bin_edges, bin_contents):
        """BinnedInterpolator constructor

        Parameters
        ----------
        grid_points: np.ndarray
            Grid points at which interpolation templates exist
        bin_edges: np.ndarray
            Edges of the data binning
        bin_content: np.ndarray
            Content of each bin in bin_edges for
            each point in grid_points. First dimesion has to correspond to number
            of grid_points, last dimension has to correspond to number of bins for
            the quantity that should be interpolated (e.g. the Migra axis for EDisp)

        Raises
        ------
        TypeError:
            When bin_edges is not a np.ndarray
        TypeError:
            When bin_content is not a np.ndarray
        ValueError:
            When number of bins in bin_edges and contents bin_contents is
            not matching
        ValueError:
            When number of histograms in bin_contents and points in grid_points
            is not matching

        Note
        ----
            Also calls pyirf.interpolation.BaseInterpolators.__call__
        """
        super().__init__(grid_points)

        if not isinstance(bin_edges, np.ndarray):
            raise TypeError("Input bin_edges is not a numpy array.")
        elif not isinstance(bin_contents, np.ndarray):
            raise TypeError("Input bin_contents is not a numpy array.")
        elif bin_contents.shape[-1] != (bin_edges.shape[0] - 1):
            raise ValueError(
                f"Shape missmatch, bin_edges ({bin_edges.shape[0] - 1} bins) "
                f"and bin_contents ({bin_contents.shape[-1]} bins) not matching."
            )
        elif self.n_points != bin_contents.shape[0]:
            raise ValueError(
                f"Shape missmatch, number of grid_points ({self.n_points}) and "
                f"number of histograms in bin_contents ({bin_contents.shape[0]}) "
                "not matching."
            )
        else:
            self.bin_edges = bin_edges
            self.bin_mids = bin_center(self.bin_edges)
            self.bin_contents = bin_contents
